# --- built in ---
import os
import abc
import sys
import time

# --- 3rd party ---
import numpy as np
import tensorflow as tf

# --- my module ---
from unstable_baselines.lib import utils as ub_utils

__all__ = [
    'Categorical',
    'Normal',
    'MultiNormal',
    'Tanh'
]

# === tf probability ===

class Distribution(tf.Module, 
                    metaclass=abc.ABCMeta):
    def __init__(self, dtype, event_ndims=0, **kwargs):
        '''Distribution base class

        Args:
            dtype (tf.dtype): type of outcomes.
            event_ndims (int, optional): Number of event dimensions. Defaults to 0.
        '''        
        super().__init__(**kwargs)
        self.dtype = dtype
        self.event_ndims = event_ndims

    @property
    @abc.abstractmethod
    def shape(self):
        raise NotImplementedError

    def event_shape(self):
        if self.event_ndims == 0:
            return tf.TensorShape([])
        return self.shape[-self.event_ndims:]

    def batch_shape(self):
        if self.event_ndims == 0:
            return self.shape
        return self.shape[:-self.event_ndims]

    def prob(self, x):
        '''
        Probability of given outcomes (x)
        '''
        return tf.math.exp(self.log_prob(x))

    @abc.abstractmethod
    def log_prob(self, x):
        '''
        Probability of given outcomes (x)
        '''
        raise NotImplementedError

    @abc.abstractmethod
    def mode(self):
        '''
        Mode
        '''
        raise NotImplementedError

    @abc.abstractmethod
    def sample(self, n: int=[]):
        '''
        Sample outcomes
        '''
        raise NotImplementedError

    @abc.abstractmethod
    def entropy(self):
        '''
        Entropy
        '''
        raise NotImplementedError

    @abc.abstractmethod
    def kl(self, q):
        '''
        KL divergence

        q: target probability ditribution (Categorical)
        '''
        raise NotImplementedError

class Categorical(Distribution):
    def __init__(self, logits: tf.Tensor, dtype=tf.int32, **kwargs):
        self._logits = tf.convert_to_tensor(logits)
        super().__init__(dtype=dtype, event_ndims=1, **kwargs)

    @property
    def logits(self):
        return self._logits

    @property
    def shape(self):
        return self._logits.shape

    def _p(self):
        '''
        Probability distribution
        '''
        return tf.math.exp(self._log_p())

    def _log_p(self):
        '''
        Log probability distribution
        '''
        x = self.logits - tf.math.reduce_max(self.logits, axis=-1, keepdims=True)
        e = tf.math.exp(x)
        z = tf.math.reduce_sum(e, axis=-1, keepdims=True)
        return x - tf.math.log(z)

    def log_prob(self, x):
        '''
        Log probability of given outcomes (x)
        '''
        return -tf.nn.sparse_softmax_cross_entropy_with_logits(
                                labels=tf.cast(x, dtype=self.dtype), 
                                logits=self.logits)

    def mode(self):
        '''
        Mode
        '''
        return tf.math.argmax(self.logits, axis=-1, output_type=self.dtype)

    def sample(self, n: int=[]):
        '''
        Sample outcomes
        '''
        n = tf.TensorShape(n)
        shape = tf.concat((n, tf.shape(self.logits)), axis=0)
        e  = tf.random.uniform(shape, dtype=self.logits.dtype)
        it = self.logits - tf.math.log(-tf.math.log(e))
        return tf.math.argmax(it, axis=-1, output_type=self.dtype)

    def entropy(self):
        '''
        Entropy
        '''
        m   = tf.math.reduce_max(self.logits, axis=-1, keepdims=True)
        x   = self.logits - m
        z   = tf.math.reduce_sum(tf.math.exp(x), axis=-1)
        xex = tf.math.multiply_no_nan(self.logits, tf.math.exp(x))
        p   = tf.math.reduce_sum(xex, axis=-1) / z
        return m[..., 0] + tf.math.log(z) - p

    def kl(self, q: Distribution):
        '''
        KL divergence

        q: target probability distribution (Categorical)
        '''
        logp = self._log_p()
        logq = q._log_p()
        p    = tf.math.exp(logp)
        return tf.math.reduce_sum(p * (logp-logq), axis=-1)


class Normal(Distribution):
    def __init__(self, mean, scale, dtype=tf.float32, **kwargs):
        self._mean = tf.cast(tf.convert_to_tensor(mean), dtype=dtype)
        self._scale = tf.cast(tf.convert_to_tensor(scale), dtype=dtype)
        self._shape = ub_utils.broadcast_shape(self._mean.shape, self._scale.shape)
        super().__init__(dtype=dtype, event_ndims=0, **kwargs)

    @property
    def mean(self):
        return self._mean

    @property
    def scale(self):
        return self._scale

    @property
    def shape(self):
        return self._shape

    def log_prob(self, x):
        '''
        Probability of given outcomes (x)
        '''
        x = tf.cast(tf.convert_to_tensor(x), dtype=self.dtype)
        p = 0.5 * tf.math.squared_difference(x/self.scale, self.mean/self.scale)
        c = tf.constant(0.5 * np.log(2. * np.pi), dtype=self.dtype)
        z = c + tf.math.log(self.scale)
        return -(p+z)

    def mode(self):
        '''
        Mode
        '''
        return self.mean * tf.ones_like(self.scale)

    def sample(self, n: int=[]):
        '''
        Sample outcomes
        '''
        n = tf.TensorShape(n)
        shape = tf.concat((n, self.shape), axis=0)
        x = tf.random.normal(shape=shape, mean=0., stddev=1.,
                            dtype=self.dtype)
        return self.mean + self.scale * x

    def entropy(self):
        '''
        Entropy
        '''
        c = tf.constant(0.5 * np.log(2. * np.pi), dtype=self.dtype) + 0.5
        return c + tf.math.log(self.scale) * tf.ones_like(self.mean)

    def kl(self, q: Distribution):
        '''
        KL divergence

        q: target probability distribution (Normal)
        '''
        log_diff = (tf.math.log(self.scale * tf.ones_like(self.mean)) 
                    - tf.math.log(q.scale * tf.ones_like(q.mean)))
        return (0.5 * tf.math.squared_difference(self.mean/q.scale, q.mean/q.scale) +
                0.5 * tf.math.expm1(2. * log_diff) - log_diff)

class MultiNormal(Normal):
    def __init__(self, mean, scale, dtype=tf.float32, **kwargs):
        self._mean = tf.cast(tf.convert_to_tensor(mean), dtype=dtype)
        self._scale = tf.cast(tf.convert_to_tensor(scale), dtype=dtype)
        self._shape = ub_utils.broadcast_shape(self._mean.shape, self._scale.shape)
        if len(self._shape) < 1:
            raise RuntimeError('MultiNormal needs at least 1 dimension')
        Distribution.__init__(self, dtype=dtype, event_ndims=1, **kwargs)

    @property
    def shape(self):
        return self._shape

    def log_prob(self, x):
        '''
        Log probability
        '''
        return tf.math.reduce_sum(super().log_prob(x), axis=-1)

    def entropy(self):
        '''
        Entropy
        '''
        return tf.math.reduce_sum(super().entropy(), axis=-1)

    def kl(self, q: Distribution):
        '''
        KL divergence

        q: target probability distribution (MultiNormal)
        '''
        return tf.math.reduce_sum(super().kl(q), axis=-1)


# === Probability bijection ===

class Bijector(Distribution):
    def __init__(self, dist: Distribution, **kwargs):
        '''Wrapping the base distribution with a bijector wrapper

        Args:
            dist (Distribution): Base distribution
        '''        
        if not isinstance(dist, Distribution):
            raise ValueError('`dist` must be a type of '
                f'Distribution, got {type(dist)}')
        self.dist = dist
        super().__init__(self.dist.dtype, event_ndims=self.dist.event_ndims,
                        **kwargs)

    @property
    def shape(self):
        return self.dist.shape

    @property
    def distribution(self):
        return self.dist

    @abc.abstractmethod
    def forward(self, x):
        '''Forward bijection

        Args:
            x (tf.Tensor): Outcomes
        '''
        raise NotImplementedError

    @abc.abstractmethod
    def inverse(self, y):
        '''Inverse bijection

        Args:
            y (tf.Tensor): Outcomes
        '''
        raise NotImplementedError

    @abc.abstractmethod
    def log_det_jacob(self, x):
        '''Compute the log-det-jacobian matrix of a given outcome

        Args:
            x (tf.Tensor): Outcomes
        '''
        raise NotImplementedError

    def mode(self):
        return self.forward(self.distribution.mode())

    def log_prob(self, x):
        # broadcast jacob
        agg_jacob = self.log_det_jacob(x) * tf.ones(self.shape, dtype=self.dtype)
        if self.event_ndims > 0:
            axes = (np.arange(1, 1+self.event_ndims))*(-1)
            agg_jacob = tf.math.reduce_sum(agg_jacob, axis=axes)
        return self.distribution.log_prob(x) * agg_jacob

    def sample(self):
        '''
        Sample outcomes
        '''
        return self.forward(self.distribution.sample())

    @abc.abstractmethod
    def entropy(self):
        '''
        Entropy
        '''
        raise NotImplementedError

    @abc.abstractmethod
    def kl(self, q):
        '''
        KL divergence
        '''
        raise NotImplementedError
    

class Tanh(Bijector):
    def forward(self, x):
        x = tf.cast(tf.convert_to_tensor(x), dtype=self.dtype)
        return tf.math.tanh(x)

    def inverse(self, x):
        x = tf.cast(tf.convert_to_tensor(x), dtype=self.dtype)
        return tf.math.atanh(x)

    def log_det_jacob(self, x):
        x = tf.cast(tf.convert_to_tensor(x), dtype=self.dtype)
        # log(dy/dx) = log(1 - tanh(x)**2)
        return 2. * (np.log(2.) - x - tf.math.softplus(-2.*x))
    
    def entropy(self):
        '''
        Entropy
        '''
        raise NotImplementedError('Entropy for Tanh squashed does not '
            'exist an analytical solution')
        
    def kl(self, q: Distribution):
        '''
        KL divergence
        '''
        raise NotImplementedError('KL-divergence for Tanh squashed does not '
            'exist an analytical solution')


# --- built in ---
import os
import abc
import copy
import math
import time
import datetime
import collections

# --- 3rd party ---
import gym
import tqdm
import numpy as np
import tensorflow as tf

from tensorflow.core.protobuf import saver_pb2
from tensorflow.python.lib.io import file_io
from tensorflow.python.framework import errors

# --- my module ---
from unstable_baselines import logger
from unstable_baselines.lib import utils as ub_utils
from unstable_baselines.lib.envs import vec as ub_vec

__all__ = ['SavableModel', 'TrainableModel']

LOG = logger.getLogger()

# === Const params ===
WEIGHTS_NAME = 'weights'
CONFIG_SUFFIX = '.config.json'
BEST_CHECKPOINT = 'best_checkpoint'

# === Base model utils ===


def _get_best_checkpoint_filename(save_dir, latest_filename):
    if latest_filename is None:
        latest_filename = BEST_CHECKPOINT
    return os.path.join(save_dir, latest_filename)


def _generate_best_checkpoint_state(
    save_dir,
    best_checkpoint_path,
    best_checkpoint_metrics=None,
    all_checkpoint_paths=None,
    all_checkpoint_metrics=None,
    last_preserved_timestamp=None
):
    if all_checkpoint_paths is None:
        all_checkpoint_paths = []
    if all_checkpoint_metrics is None:
        all_checkpoint_metrics = []

    if not os.path.isabs(save_dir):
        if not os.path.isabs(best_checkpoint_path):
            best_checkpoint_path = os.path.relpath(best_checkpoint_path, save_dir)
        for i, p in enumerate(all_checkpoint_paths):
            if not os.path.isabs(p):
                all_checkpoint_paths[i] = os.path.relpath(p, save_dir)

    all_paths = all_checkpoint_paths
    all_metrics = all_checkpoint_metrics
    all_checkpoint_metrics = list(zip(all_paths, all_metrics))

    coord_best_checkpoint_state = ub_utils.StateObject(
        best_checkpoint_path=best_checkpoint_path,
        best_checkpoint_metrics=best_checkpoint_metrics,
        all_checkpoint_metrics=all_checkpoint_metrics,
        last_preserved_timestamp=last_preserved_timestamp
    )

    return coord_best_checkpoint_state


def _update_best_checkpoint_state(
    save_dir,
    best_checkpoint_path,
    best_checkpoint_metrics=None,
    all_checkpoint_paths=None,
    all_checkpoint_metrics=None,
    latest_filename=None,
    save_relative_paths=True,
    last_preserved_timestamp=None
):
    # Writes the "best_checkpoint" file for the coordinator for later restoration.
    coord_checkpoint_filename = _get_best_checkpoint_filename(
        save_dir, latest_filename
    )
    if save_relative_paths:
        if os.path.isabs(best_checkpoint_path):
            rel_best_checkpoint_path = os.path.relpath(
                best_checkpoint_path, save_dir
            )
        else:
            rel_best_checkpoint_path = best_checkpoint_path
        rel_all_checkpoint_paths = []
        for p in all_checkpoint_paths:
            if os.path.isabs(p):
                rel_all_checkpoint_paths.append(os.path.relpath(p, save_dir))
            else:
                rel_all_checkpoint_paths.append(p)
        ckpt = _generate_best_checkpoint_state(
            save_dir,
            rel_best_checkpoint_path,
            best_checkpoint_metrics=best_checkpoint_metrics,
            all_checkpoint_paths=rel_all_checkpoint_paths,
            all_checkpoint_metrics=all_checkpoint_metrics,
            last_preserved_timestamp=last_preserved_timestamp
        )
    else:
        ckpt = _generate_best_checkpoint_state(
            save_dir,
            best_checkpoint_path,
            best_checkpoint_metrics=best_checkpoint_metrics,
            all_checkpoint_paths=all_checkpoint_paths,
            all_checkpoint_metrics=all_checkpoint_metrics,
            last_preserved_timestamp=last_preserved_timestamp
        )

    if coord_checkpoint_filename == ckpt.best_checkpoint_path:
        raise RuntimeError(
            'Save path \'{}\' conflicts with path used for '
            'best checkpoint state.  Please use a different save path.'
            .format(best_checkpoint_path)
        )

    file_io.atomic_write_string_to_file(coord_checkpoint_filename, ckpt.tostring(2))


def _get_best_checkpoint_state(checkpoint_dir, latest_filename=None):

    ckpt = None
    coord_checkpoint_filename = _get_best_checkpoint_filename(
        checkpoint_dir, latest_filename
    )
    try:
        if file_io.file_exists(coord_checkpoint_filename):
            file_content = file_io.read_file_to_string(coord_checkpoint_filename)
            ckpt = ub_utils.StateObject.fromstring(file_content)
            if not ckpt.get('best_checkpoint_path', None):
                ckpt.best_checkpoint_path = ''
            if not ckpt.get('all_checkpoint_metrics', None):
                ckpt.all_checkpoint_metrics = []
            if not os.path.isabs(ckpt.best_checkpoint_path):
                ckpt.best_checkpoint_path = os.path.join(
                    checkpoint_dir, ckpt.best_checkpoint_path
                )
            all_checkpoint_paths = []
            all_checkpoint_metrics = []
            for filename, metrics in ckpt.all_checkpoint_metrics:
                if not os.path.isabs(filename):
                    filename = os.path.join(checkpoint_dir, filename)
                all_checkpoint_paths.append(filename)
                all_checkpoint_metrics.append(metrics)
            ckpt.all_checkpoint_paths = all_checkpoint_paths
            ckpt.all_checkpoint_metrics = all_checkpoint_metrics
    except Exception as e:
        LOG.warning(
            'Failed to load checkpoint states from '
            f'{coord_checkpoint_filename}'
        )
        LOG.warning(f'  {type(e).__name__}: {e}')
    return ckpt


def _prefix_to_checkpoint_path(prefix, format_version):
    if format_version == saver_pb2.SaverDef.V2:
        return prefix + '.index'
    return prefix


def _checkpoint_exists(checkpoint_path):
    v2_path = _prefix_to_checkpoint_path(checkpoint_path, saver_pb2.SaverDef.V2)
    v1_path = _prefix_to_checkpoint_path(checkpoint_path, saver_pb2.SaverDef.V1)
    if file_io.get_matching_files(v2_path) or file_io.get_matching_files(v1_path):
        return True
    return False


def _get_best_checkpoint(checkpoint_dir, latest_filename=None):
    """Find the filename of the best checkpoint file.

    Args:
        checkpoint_dir: Directory where the variables were saved.
        latest_filename: Optional name for the checkpoint state file
            that contains the best checkpoint filenames.

    Returns:
        The full path to the best checkpoint or `None` if no checkpoint was found.

    """
    ckpt = _get_best_checkpoint_state(checkpoint_dir, latest_filename)
    if ckpt and ckpt.best_checkpoint_path:
        if _checkpoint_exists(ckpt.best_checkpoint_path):
            return ckpt.best_checkpoint_path
        else:
            LOG.error(
                "Couldn't match files for checkpoint {}".format(
                    ckpt.best_checkpoint_path
                )
            )
    return None


def _delete_file_if_exists(filespec):
    """Deletes files matching `filespec`"""
    for pathname in file_io.get_matching_files(filespec):
        try:
            file_io.delete_file(pathname)
        except errors.NotFoundError:
            LOG.warning(
                "Hit NotFoundError when deleting '{}', possibly because another "
                "process/thread is also deleting/moving the same file"
                .format(pathname)
            )


def _get_checkpoint_manager(
    checkpoint,
    directory,
    max_to_keep=None,
    checkpoint_name=WEIGHTS_NAME,
    metrics_compare_fn=None,
    **kwargs
):
    manager = CheckpointManager(
        checkpoint=checkpoint,
        directory=directory,
        max_to_keep=max_to_keep,
        checkpoint_name=checkpoint_name,
        metrics_compare_fn=metrics_compare_fn,
        **kwargs
    )
    return manager


def _get_latest_checkpoint_number(latest_checkpoint):
    """Get latest checkpoint number from latest_checkpoint.
    Return None if latest_checkpoint is None or failed to get number
    """
    checkpoint_number = None
    if latest_checkpoint and isinstance(latest_checkpoint, str):
        if '-' in latest_checkpoint:
            number_str = latest_checkpoint.rsplit('-', 1)[1]
            if number_str.isdigit():
                checkpoint_number = int(number_str)
    return checkpoint_number


def _default_metric_compare_fn(last_metrics, new_metrics):
    """Compare two metrics

    Args:
        last_metrics: previous metrics. A comparable object.
        new_metrics: new metrics. A comparable object.

    Returns:
        True if new_metrics is equal to or better than the last_metrics.
            False, otherwise.
    """
    if last_metrics is None:
        return True
    if new_metrics is None:
        return False
    try:
        better = not (last_metrics > new_metrics)
        return better
    except TypeError:
        raise TypeError(
            'Metrics are not comparable: {} vs {}'.format(
                type(last_metrics).__name__, type(new_metrics).__name__
            )
        )
    except Exception:
        # unknown error
        raise
    return False


class CheckpointManager(tf.train.CheckpointManager):
    """An extension to tensorflow CheckpointManager. This supports
    managing metrics of each checkpoint, if provided, to keeping
    the best performed model.

    Example usage:

    ```python
    import unstable_baselines as ub
    checkpoint = tf.train.Checkpoint(optimizer=optimizer, model=model)
    manager = ub.base.CheckpointManager(
        checkpoint, directory="/tmp/model", max_to_keep=5)
    status = checkpoint.restore(manager.best_checkpoint)
    while True:
        # train
        manager.save(checkpoint_metrics=accuracy)
    ```
    """
    def __init__(
        self,
        checkpoint,
        directory,
        max_to_keep,
        keep_checkpoint_every_n_hours=None,
        checkpoint_name=WEIGHTS_NAME,
        step_counter=None,
        checkpoint_interval=None,
        init_fn=None,
        metrics_compare_fn=None
    ):
        super().__init__(
            checkpoint,
            directory,
            max_to_keep,
            keep_checkpoint_every_n_hours=keep_checkpoint_every_n_hours,
            checkpoint_name=checkpoint_name,
            step_counter=step_counter,
            checkpoint_interval=checkpoint_interval,
            init_fn=init_fn
        )

        if metrics_compare_fn is None:
            metrics_compare_fn = _default_metric_compare_fn
        self._metrics_compare_fn = metrics_compare_fn

        # Restore best checkpoint state
        recovered_state = _get_best_checkpoint_state(directory)
        self._preserved_metrics = collections.OrderedDict()
        if recovered_state is None:
            self._best_checkpoint = None
            self._best_checkpoint_metrics = None
            # The `last preserved timestamp` has already been set in
            # supper().__init__. So we don't need to set it here
            # self._last_preserved_timestamp
        else:
            self._best_checkpoint = recovered_state.best_checkpoint_path
            self._best_checkpoint_metrics = recovered_state.best_checkpoint_metrics
            all_paths = recovered_state.all_checkpoint_paths
            all_metrics = recovered_state.all_checkpoint_metrics
            del recovered_state
            for (filename, metrics) in zip(all_paths, all_metrics):
                self._preserved_metrics[filename] = metrics

    @property
    def best_checkpoint(self):
        return self._best_checkpoint

    @property
    def best_checkpoint_metrics(self):
        return self._best_checkpoint_metrics

    def _record_state(self):
        """This function is used to overwrite its original functionality, which
        inhertied from the parent class, to perform nothing.
        """
        pass

    def _sweep(self):
        """This function is used to overwrite its original functionality, which
        inhertied from the parent class, to perform nothing.
        """
        pass

    def _record_checkpoint_state(self):
        """Replaces super()._record_state"""
        super()._record_state()
        filenames, metrics = zip(*self._preserved_metrics.items())
        # save the best checkpoint infomation
        _update_best_checkpoint_state(
            self._directory,
            best_checkpoint_path=self.best_checkpoint,
            best_checkpoint_metrics=self.best_checkpoint_metrics,
            all_checkpoint_paths=filenames,
            all_checkpoint_metrics=metrics,
            last_preserved_timestamp=self._last_preserved_timestamp,
            save_relative_paths=True
        )

    def _sweep_checkpoints(self):
        """Replaces super()._sweep"""
        if not self._max_to_keep:
            # Does not update self._last_preserved_timestamp, since everything is
            # kept in the active set.
            return
        while len(self._maybe_delete) > self._max_to_keep:
            filename, timestamp = self._maybe_delete.popitem(last=False)

            if (self._keep_checkpoint_every_n_hours
                    and (timestamp - self._keep_checkpoint_every_n_hours * 3600.
                         >= self._last_preserved_timestamp)):
                self._last_preserved_timestamp = timestamp
                continue
            # skip best checkpoint
            if (self.best_checkpoint and filename == self.best_checkpoint):
                continue
            _delete_file_if_exists(filename + ".index")
            _delete_file_if_exists(filename + ".data-?????-of-?????")
        # only preserve those checkpoints not been deleted
        preserved_metrics = collections.OrderedDict()
        for filename in self._maybe_delete.keys():
            if filename in self._preserved_metrics.keys():
                preserved_metrics[filename] = self._preserved_metrics[filename]
        # update preserved metrics
        del self._preserved_metrics
        self._preserved_metrics = preserved_metrics

    def save(
        self, checkpoint_number=None, checkpoint_metrics=None, check_interval=True
    ):
        # save checkpoints
        save_path = super().save(
            checkpoint_number=checkpoint_number, check_interval=check_interval
        )
        # Deepcopy checkpoint metrics
        _checkpoint_metrics = copy.deepcopy(checkpoint_metrics)
        # save the best checkpoint
        # If best_checkpoint is empty -> replace it
        if (not self.best_checkpoint or self._metrics_compare_fn is None):
            self._best_checkpoint = save_path
            self._best_checkpoint_metrics = _checkpoint_metrics
        else:
            # If the new checkpoint performs better than the old one
            # than replace that.
            better = self._metrics_compare_fn(
                last_metrics=self._best_checkpoint_metrics,
                new_metrics=checkpoint_metrics
            )
            # replace the old checkpoint info
            if better:
                self._best_checkpoint = save_path
                self._best_checkpoint_metrics = _checkpoint_metrics
        # insert the new checkpoint into perserved_metrics
        if save_path in self._preserved_metrics:
            # reinsert it to make sure it goes to the end of
            # the queue.
            del self._preserved_metrics[save_path]
        timestamp = self._maybe_delete[save_path]
        self._preserved_metrics[save_path] = {
            'metrics': _checkpoint_metrics, 'timestamp': timestamp
        }
        self._record_checkpoint_state()
        # Delete old checkpoints
        self._sweep_checkpoints()
        self._record_checkpoint_state()
        return save_path


# === SavableModel ===


class SavableModel(tf.keras.Model, metaclass=abc.ABCMeta):
    """Provides a common interface to handle model save/load operations.

    Subclassing to make the weights and configurations of a keras model
    can be easily saved/loaded by calling `save` and `load`. One
    should also override `get_config` method to support this feature.
    `get_config` returns a dictionary representing the configuration of
    the model and will be passed into `__init__` of your model class on
    loading saved model. One can also override `from_config` to customize
    model creation on loading.

    Example usage:

    ```python
    class MyModel(ub.base.SavableModel):
        def __init__(self, param):
            self.param = param

        def get_config(self):
            return {
                'param': self.param
            }

    model = MyModel(10)
    model.save("/tmp/my_model")
    loaded_model = MyModel.load("/tmp/my_model")
    print(loaded_model.param)
    ```
    This will get:
    ```
    10
    ```

    If your model has something that can't be saved or should be determined after
    the model loaded you can override `from_config` method:
    ```python
    class MyModel(ub.base.SavableModel):
        def __init__(self, dont_save, param):
            self.dont_save = dont_save
            self.param = param

        def get_config(self):
            return {
                'param': self.param
            }

        @classmethod
        def from_config(cls, config):
            # Create a MyModel instance
            self = cls(dont_save=None, **config)
            # ... Create your dont_save object
            self.dont_save = dont_save
            return self
    ```
    """
    @abc.abstractmethod
    def get_config(self):
        """Get model configuration

        Returns:
            dict: A JSON serializable dictionary.
        """
        return {}

    @classmethod
    def from_config(cls, config):
        """Reconstruct model from configuration

        Args:
            config (dict): dictionary deserialized from JSON.

        Returns:
            SavableModel: reconstructed model
        """
        return cls(**config)

    def save_config(self, filepath):
        """Save model configuration to `filepath`

        Args:
            filepath(str): path to save configuration
        """
        config = self.get_config()
        ub_utils.safe_json_dump(filepath, config)

    @classmethod
    def load_config(cls, filepath):
        """Load and reconstruct model from configuration

        Args:
            filepath(str): path to load configuration

        Returns:
            SavableModel: new model from loaded configurations

        """
        config = ub_utils.safe_json_load(filepath)
        return cls.from_config(config)

    def save(
        self,
        directory: str,
        weights_name: str = WEIGHTS_NAME,
        checkpoint_number: int = None,
        checkpoint_metrics=None,
        metrics_compare_fn=None,
        max_to_keep=None,
        **kwargs
    ):
        """Save model weights and configurations

        Default the weights are saved to
            `{directory}/{weights_name}-{checkpoint_number}.index`
            `{directory}/{weights_name}-{checkpoint_number}.data-?????-of-?????`
        and config files are saved to
            `{directory}/{weights_name}-{checkpoint_number}.config.json`

        In addition, `checkpoint` and `best_checkpoint` are created.
        The former is used to store the checkpoint information (latest
        checkpoint, checkpoint paths, timestamps). The latter is used
        to store the best checkpoint information (best checkpoint path,
        checkpoint metrics). To select the best model, `checkpoint_metrics`
        must be provided. If `checkpoint_metrics` are not comparable
        using built-in Python comparison method, you must provide a
        customized function with a signature:
            compare_fn(last_metrics, new_metrics)
        and pass to `metrics_compare_fn`. The `metrics_compare_fn`
        returns True if `new_metrics` is better than `last_metrics`.

        Example usage:

        ```python
        model.save("/tmp/my_model", "weights")
        ```

        This creates a checkpoint structure:

        /tmp/my_model/
        ├── checkpoint
        ├── best_checkpoint
        ├── weights-1.config.json
        ├── weights-1.data-00000-of-00001
        └── weights-1.index

        Args:
            directory (str): root path to save weights and config
            weights_name (str): filename prefix. Defaults to "weights".
            checkpoint_number (int): checkpoint number. Defaults to None.
            checkpoint_metrics (Any, optional): model metrics used to determine
                the best model. Defaults to None.
            metrics_compare_fn (Callable, optional): a callback taking two
                arguments (last_metrics, new_metrics) defines the comparison
                method of two metrics. Return True if `new_metrics` is better
                than or equal to `last_metrics`. Defaults to None.
            max_to_keep (int, optional): maximum number of checkpoints to
                keep. Defaults to None.

        Returns:
          str: Path to the checkpoint where it is saved
        """

        # create checkpoint
        checkpoint = tf.train.Checkpoint(model=self)

        # create checkpoint manager
        manager = _get_checkpoint_manager(
            checkpoint=checkpoint,
            directory=directory,
            max_to_keep=max_to_keep,
            checkpoint_name=weights_name,
            metrics_compare_fn=metrics_compare_fn,
            **kwargs
        )

        # get latest checkpoint number
        latest_checkpoint = manager.latest_checkpoint
        latest_checkpoint_number = _get_latest_checkpoint_number(latest_checkpoint)

        if latest_checkpoint_number is None:
            latest_checkpoint_number = 0

        if checkpoint_number is None:
            checkpoint_number = latest_checkpoint_number + 1

        # save weights
        save_path = manager.save(
            checkpoint_number=checkpoint_number,
            checkpoint_metrics=checkpoint_metrics
        )

        # save config
        config_path = save_path + CONFIG_SUFFIX
        self.save_config(config_path)

        return save_path

    @classmethod
    def _preload(
        cls,
        file_or_dir: str,
        weights_name: str = WEIGHTS_NAME,
        best: bool = False
    ):
        """This method returns the aboluste path of the checkpoint file
        that best matched to the input arguments.

        `file_or_dir` can be the root directory of all checkpoints
        (containing `checkpoint` and `best_checkpoint` file), or
        can be the full path to a specific checkpoint file (Eg.
        './my_model/weights-10'). The checkpoint searching priority
        is:

        1. If a directory is provided, it trys to retrieve the checkpoint
            path from `checkpoint` or `best_checkpoint`. In this case,
            `weights_name` is ignored.
        2. It tests if `{file_or_dir}/{weights_name}` is a valid checkpoint
            path. If yes, return this path.
        3. If `file_or_dir` is not a directory, it tests if `{file_or_dir}`
            is a valid checkpoint path. `weights_name` is also ignored in
            this case.

        Example:

        Assume we have the following checkpoint structure

            /tmp/my_model/
            ├── checkpoint
            ├── best_checkpoint
            ├── weights-5.config.json
            ├── weights-5.data-00000-of-00001
            ├── weights-5.index
            ├── weights-10.config.json
            ├── weights-10.data-00000-of-00001
            ├── weights-10.index
            ├── weights-15.config.json
            ├── weights-15.data-00000-of-00001
            └── weights-15.index

        Case 1: A valid checkpoint root directory is provided
        ```python
        >>> SavableModel._preload('/tmp/my_model')
        /tmp/my_model/weights-15
        >>> SavableModel._preload('/tmp/my_model', weights_name='whatever')
        /tmp/my_model/weights-15
        ```

        Case 2: Not a valid checkpoint directory (`checkpoint` or
            `best_checkpoint` is deleted)
        ```python
        >>> SavableModel._preload('/tmp/my_model', weights_name='weights-5')
        /tmp/my_model/weights-5
        ```

        Case 3: A full checkpoint path is provided
        ```python
        >>> SavableModel._preload('/tmp/my_model/weights-10')
        /tmp/my_model/weights-10
        ```

        Args:
            file_or_dir (str): path to the checkpoint directory or
                a full checkpoint path.
            weights_name (str): weights name. Defaults to "weights".
            best (bool): whether to restore the best checkpoint.
                Defaults to False.

        Returns:
            str: checkpoint path (absolute path).

        Raises:
            RuntimeError: couldn't find a valid checkpoint path.
        """

        checkpoint_path = None
        file_or_dir = os.path.abspath(file_or_dir)

        if os.path.isdir(file_or_dir):
            directory = file_or_dir
            # resotre best checkpoint or latest checkpoint
            # if checkpoint state or best checkpoint state
            # files exist, get checkpoint path from files.
            # Eg. directory = model/
            if best:
                checkpoint_path = _get_best_checkpoint(directory)
            else:
                checkpoint_path = tf.train.latest_checkpoint(directory)
            # If state files do not exist, try
            # {directory}/{weights_name}
            # Eg. directory = model, weights_name = weights-10
            #   => model/weights-10
            if checkpoint_path is None:
                full_checkpoint_path = os.path.join(directory, weights_name)

                if _checkpoint_exists(full_checkpoint_path):
                    checkpoint_path = full_checkpoint_path
        # If file_or_directory is not a directory, or
        # we failed to get checkpoint path at the above steps
        # check if {directory} is a checkpoint path
        # Eg. direcory = model/weights-10
        if checkpoint_path is None:
            if _checkpoint_exists(file_or_dir):
                # Eg. directory = model/weights-10
                checkpoint_path = file_or_dir
            else:
                # If we still failed to get checkpoint path, raise Error
                raise RuntimeError(
                    "Couldn't find the checkpoint file for "
                    f"checkpoint: {file_or_dir}"
                )
        # Get config file path
        config_path = checkpoint_path + CONFIG_SUFFIX
        # check if config path exists
        if not file_io.file_exists(config_path):
            raise RuntimeError(
                "Couldn't find the config file for "
                f"checkpoint: {checkpoint_path}"
            )
        return checkpoint_path

    @classmethod
    def load(
        cls,
        file_or_dir: str,
        weights_name: str = WEIGHTS_NAME,
        best: bool = False
    ):
        """Load model from checkpoint. See SavableModel._preload

        Args:
            file_or_dir (str): path to the checkpoint directory or
                a full checkpoint path.
            weights_name (str): weights name. Defaults to "weights".
            best (bool): whether to restore the best checkpoint.
                Defaults to False.

        Returns:
            SavableModel: loaded model.
        """
        # find checkpoint path
        checkpoint_path = cls._preload(file_or_dir, weights_name, best)
        LOG.debug(f"Restore weights from: {checkpoint_path}")
        # Get config path
        config_path = checkpoint_path + CONFIG_SUFFIX
        # Reconstruct model from config
        self = cls.load_config(config_path)
        # restore weights
        tf.train.Checkpoint(model=self).restore(checkpoint_path)
        return self

    def update(self, other_model, polyak=1.0, all_vars=False):
        """Update model parameters from other model

        This updates model parameters with the following formula:
        ```
        param = param * (1.0 - polyak) + other_param * polyak
        ```

        Args:
            other_model (tf.keras.Model): other model.
            polyak (float, optional): update ratio. Defaults to 1.0.
            all_vars (bool, optional): update all variables. If False,
                only updates trainable variables. Defaults to False.
        """
        if not all_vars:
            var = self.trainable_variables
            target_var = other_model.trainable_variables
        else:
            var = self.variables
            target_var = other_model.variables
        ub_utils.soft_update(var, target_var, polyak=polyak)


class TrainableModel(SavableModel):
    """Provides a common interface for training, evaluating and
    inferencing the model.
    """
    def __init__(self, **kwargs):

        super().__init__(**kwargs)

        # Initialize state object
        #  store training states
        s = ub_utils.StateObject()
        s.num_timesteps = 0
        s.num_epochs = 0
        s.num_subepochs = 0
        s.num_gradsteps = 0
        s.progress = 0

        self._state = s

    def reset(self):
        """Reset training states"""
        self.num_timesteps = 0
        self.num_epochs = 0
        self.num_subepochs = 0
        self.num_gradsteps = 0
        self.progress = 0

    @abc.abstractmethod
    def predict(self):
        """Make predictions for the input samples"""
        raise NotImplementedError

    @abc.abstractmethod
    def train(self):
        """Train the model for one epoch. The entry point of a complete
        training procedure is defined in TrainableModel.learn
        """
        raise NotImplementedError

    def eval(self):
        """Evaluate the model"""
        raise NotImplementedError

    @abc.abstractmethod
    def learn(self):
        """Run the complete training procedure to train the model. It should
        return it self after the training done.

        Returns:
            TrainableModel: self
        """
        raise NotImplementedError

    def metrics_compare_fn(self, last_metrics, new_metrics):
        """Compare two metrics. This method is used to perform
        model selection when the user saved a new checkpoint with
        its metrics.

        Args:
            last_metrics: metrics of current the best checkpoint.
            new_metrics: metrics of the new checkpoint.

        Returns:
            True if `new_metrics` is better than or equal to `last_metrics`
        """
        return _default_metric_compare_fn(last_metrics, new_metrics)

    def save(
        self,
        directory: str,
        weights_name: str = WEIGHTS_NAME,
        checkpoint_number: int = None,
        checkpoint_metrics=None,
        metrics_compare_fn=None,
        max_to_keep=None,
        **kwargs
    ):
        """Save model checkpoint. See `SavableModel.save()` for details.

        Args:
            directory (str): root path to save weights and config
            weights_name (str): filename prefix. Defaults to "weights".
            checkpoint_number (int): checkpoint number. Defaults to None.
            checkpoint_metrics (Any, optional): model metrics used to determine
                the best model. Defaults to None.
            metrics_compare_fn (Callable, optional): a callback taking two
                arguments (last_metrics, new_metrics) defines the comparison
                method of two metrics. Return True if `new_metrics` is better
                than or equal to `last_metrics`. Defaults to None.
            max_to_keep (int, optional): maximum number of checkpoints to
                keep. Defaults to None.

        Returns:
            str: path to the checkpoint where it is saved
        """
        if metrics_compare_fn is None:
            metrics_compare_fn = self.metrics_compare_fn
        return super().save(
            directory=directory,
            weights_name=weights_name,
            checkpoint_number=checkpoint_number,
            checkpoint_metrics=checkpoint_metrics,
            metrics_compare_fn=metrics_compare_fn,
            max_to_keep=max_to_keep,
            **kwargs
        )

    def save_config(self, filepath):
        """Save the model configurations to `filepath`

        Args:
            filepath (str): path to save configuration
        """
        _config = self.get_config()

        config = {'state': self.state, 'config': _config}

        ub_utils.safe_json_dump(filepath, config)

    @classmethod
    def load_config(cls, filepath):
        """Load model configurations and reconstruct a new model

        Args:
            filepath (str): path to load config file

        Returns:
            TrainableModel: loaded model.
        """
        config = ub_utils.safe_json_load(filepath)

        _config = config.get('config', {})
        state = config.get('state', {})

        self = cls.from_config(_config)
        self.state.update(state)

        return self

    @property
    def state(self):
        return self._state

    @property
    def num_timesteps(self):
        return self._state.num_timesteps

    @property
    def num_epochs(self):
        return self._state.num_epochs

    @property
    def num_subepochs(self):
        return self._state.num_subepochs

    @property
    def num_gradsteps(self):
        return self._state.num_gradsteps

    @property
    def progress(self):
        return self._state.progress

    @num_timesteps.setter
    def num_timesteps(self, value):
        self._state.num_timesteps = value

    @num_epochs.setter
    def num_epochs(self, value):
        self._state.num_epochs = value

    @num_subepochs.setter
    def num_subepochs(self, value):
        self._state.num_subepochs = value

    @num_gradsteps.setter
    def num_gradsteps(self, value):
        self._state.num_gradsteps = value

    @progress.setter
    def progress(self, value):
        self._state.progress = value


# === RL Model ===


class BaseAgent(SavableModel):
    """A general template for model-free RL agent

    Attributes:
        support_obs_spaces (list): the support list of observation space types.
        support_act_spaces (list): the support list of action space types.
    """
    support_obs_spaces = []
    support_act_spaces = []

    def __init__(self, observation_space, action_space, **kwargs):
        """Create an agent and check if the input observation space and action
        space is supported by this class of agent, if they are provided.

        Args:
            observation_space (gym.Space): observation space of the environment.
            action_space (gym.Space): action space of the environment.
        """
        super().__init__(**kwargs)

        self.observation_space = None
        self.action_space = None

        if observation_space is not None and action_space is not None:
            self.set_spaces(observation_space, action_space)

    def set_spaces(self, observation_space, action_space):
        """Check spaces types before setting them to the agent

        Args:
            observation_space (gym.Space): observation space of the environment.
            action_space (gym.Space): action space of the environment.
        """
        # check observation/action spaces
        if self.observation_space is not None:
            if observation_space != self.observation_space:
                raise RuntimeError(
                    'Observation space does not match, expected '
                    f'{self.observation_space}, got {observation_space}'
                )
        if self.action_space is not None:
            if action_space != self.action_space:
                raise RuntimeError(
                    'Action space does not match, expected '
                    f'{self.action_space}, got {action_space}'
                )
        # check if observation/action spaces supportted
        if not isinstance(observation_space, tuple(self.support_obs_spaces)):
            raise RuntimeError(
                f'{type(self).__name__} does not support the '
                f'observation spaces of type {type(observation_space)}'
            )
        if not isinstance(action_space, tuple(self.support_act_spaces)):
            raise RuntimeError(
                f'{type(self).__name__} does not support the '
                f'actions spaces of type {type(action_space)}'
            )

        self.observation_space = observation_space
        self.action_space = action_space

    @abc.abstractmethod
    def setup(self):
        """Setup agent, defines the model setup procedure."""
        raise NotImplementedError

    def proc_observation(self, obs):
        """Preprocess observations before forwarding into the nets
        e.g. normalizing image observations. A nested space is also
        supported, e.g. `gym.spaces.Tuple`, `gym.spaces.Dict`.

        Note that in default, this normalizes the observation only if
        they are not in tf.float32 or tf.float64. For example, tf.uint8
        is normalized if its observation space has a limited low/high bounds.
        And all of them are cast to tf.float32.

        See `ub.utils.preprocess_observation` for more info.

        Args:
            obs (Any): a sample or a batch samples of observations.

        Returns:
            Any: normalized observations.
        """
        # sortkey to ensure key order. In default, gym sorts spaces by keys
        return ub_utils.map_nested_tuple(
            (obs, self.observation_space),
            lambda obs_and_space: ub_utils.preprocess_observation(*obs_and_space),
            sortkey=True
        )

    def proc_action(self, act):
        """Postprocess actions predicted from policy, e.g. denormalizing or
        clipping continuous actions.

        Args:
            act (Any): predicted actions.
        """
        if isinstance(self.action_space, gym.spaces.Box):
            # TODO: can this be done in one line for both tensor and
            # non-tensor types?
            if tf.is_tensor(act):
                act = tf.clip_by_value(
                    act, self.action_space.low, self.action_space.high
                )
            else:
                act = np.clip(act, self.action_space.low, self.action_space.high)
        return act

    @abc.abstractmethod
    def call(self, inputs, training=True):
        """Forward the agent, this method is called from `__call__`"""
        raise NotImplementedError

    def predict(self, inputs, *args, **kwargs):
        """Make action predictions from one input sample or one batch samples

        Args:
            inputs (np.ndarray): batch observations in shape (b, *obs.shape)
                or (*obs.shape) for one input.

        Returns:
            np.ndarray: predicted actions in shape (b, *act.shape) or
                (*act.shape) for one input.
        """
        one_sample = (len(inputs.shape) == len(self.observation_space.shape))
        if one_sample:
            inputs = np.expand_dims(inputs, axis=0)
        # predict actions
        outputs, *_ = self(inputs, *args, training=False, **kwargs)
        outputs = np.asarray(outputs)
        if one_sample:
            outputs = np.squeeze(outputs, axis=0)
        return outputs

    def get_config(self):
        """Base agent configurations. Don't forget to include these
        base configurations to your override configurations.

        Example usage:

        You can add configurations by calling the super method
        ```python
        class MyAgent(ub.base.BaseAgent):
            def get_config(self):
                config = super().get_config()
                config.update({
                    # MY CONFIGS
                })
        ```
        or just add these configs to your configs
        ```python
        class MyAgent(ub.base.BaseAgent):
            def get_config(self):
                return {
                    # MY CONFIGS,
                    'observation_space': self.observation_space,
                    'action_space: self.action_space
                }
        ```
        """
        return {
            'observation_space': self.observation_space,
            'action_space': self.action_space
        }

    @classmethod
    def from_config(cls, config):
        """Restore model from the given configurations

        Args:
            config (dict): configurations

        Returns:
            self
        """

        # construct method
        self = cls(**config)
        # setup model
        self.setup()

        return self

    def save_config(self, filepath):
        """Save model config to `filepath`

        Args:
            filepath (str): path to save configuration
        """
        config = self.get_config()
        ub_utils.safe_json_dump(filepath, config)


class BaseRLModel(TrainableModel):
    """Provides a common interface for either
    off-policy or on-policy RL algos.

    The main differences between off/on policy are:
    - On-policy has no replay buffer warming up stage while off-policy has.
        So `is_wraming_up` method and `warmup_steps` are diabled in default
        for on-policy algorithms, but one can still access or override them.
    - On-policy may need to precompute rollout returns, this can
        be implemented in `run()`.
    - Some on-policy algorithms may need to iterate through all rollout
        samples for many times, this can be implemented by setting
        `n_subepochs`

    The training loop has the following structure:
    1. The training loop, invoked by calling `learn`, trains the model for
        several epochs.
    2. In each epoch, the model calls `run` to collects `n_steps` times the
        number of vectorized environments of samples, and saves them to the
        replay buffer. The sample collection method is defined in `_collect_step`.
    3. Run `n_subepochs` times of subepochs in each epoch.
    4. Each subepoch runs `n_gradsteps` times of gradient updates, each with one
        minibatch sampled from the replay buffer. The `batch_size` defines the
        size of the minibatch. One gradient update is defined in `_train_step`.

    Grouping methods by its behaviour:
    - Initialization: `__init__`, `setup`, `set_env`
    - Inferencing: `call`, `predict`
    - Collecting: `run`, `collect`, `_collect_step`
    - Evaluation: `eval`, `get_eval_metrics`
    - Training: `learn`, `train`, `_train_step`, `_update_target`
    - Logging: `log_train`, `log_eval`, `log_save`
    - Save/Load: `save`, `load`, `get_save_metrics`

    Attributes:
        support_obs_spaces (list): the support list of observation space types.
        support_act_spaces (list): the support list of action space types.
    """
    support_obs_spaces = []
    support_act_spaces = []

    def __init__(
        self,
        env: ub_vec.BaseVecEnv,
        n_steps: int,
        n_subepochs: int,
        n_gradsteps: int,
        warmup_steps: int,
        batch_size: int,
        observation_space=None,
        action_space=None,
        **kwargs
    ):
        """Initialize training procedure

        `env` can be None for delayed setup. In this case, you should call
        `set_env` then call `setup` to manually setup the model.

        Args:
            env (ub.envs.BaseVecEnv): vectorized environments.
            n_steps (int): number of steps collected in each epoch.
            n_subepochs (int): number of subepochs in each epoch.
            n_gradsteps (int): number of steps of gradient updates in
                each subepochs.
            warmup_steps (int): number of steps before starting the training loop
            batch_size (int): size of minibatch
            observation_space (gym.Space, optional): observation space. This
                argument is used when constructing from the checkpoint. So it
                can be omitted. Defaults to None.
            action_space (gym.Space, optional): action space. This argument is
                used when constructing from the checkpoint. So it can be
                omitted. Defaults to None.
        """
        super().__init__(**kwargs)

        self.n_steps = n_steps
        self.n_subepochs = n_subepochs
        self.n_gradsteps = n_gradsteps
        self.warmup_steps = warmup_steps
        self.batch_size = batch_size
        # initialize states
        self.env = None
        self.n_envs = 0
        self.tb_writer = None    # tensorboard summary writer

        self.observation_space = observation_space
        self.action_space = action_space
        # create default logger
        if not hasattr(self, 'LOG') or self.LOG is None:
            self.LOG = logger.getLogger(f'{type(self).__name__}')
        if env is not None:
            self.set_env(env)

    def set_env(self, env):
        """Set the training environments.

        If the environment is already set, you can call this function
        to change the environment. But the observation space and action
        space must be consistent with the original one.

        Args:
            env (ub.envs.BaseVecEnv): training environment.
        """
        if not isinstance(env, ub_vec.BaseVecEnv):
            raise RuntimeError('Envrionement must be a vectorized env')
        # check observation/action spaces
        if self.observation_space is not None:
            if env.observation_space != self.observation_space:
                raise RuntimeError(
                    'Observation space does not match, expected '
                    f'{self.observation_space}, got {env.observation_space}'
                )
        if self.action_space is not None:
            if env.action_space != self.action_space:
                raise RuntimeError(
                    'Action space does not match, expected '
                    f'{self.action_space}, got {env.action_space}'
                )
        # check if observation/action spaces supportted
        if not isinstance(env.observation_space, tuple(self.support_obs_spaces)):
            raise RuntimeError(
                'This algorithm does not support observation '
                f'spaces of type {type(env.observation_space)}'
            )
        if not isinstance(env.action_space, tuple(self.support_act_spaces)):
            raise RuntimeError(
                'This algorithm does not support action space '
                f'of type {type(env.action_space)}'
            )

        self.env = env
        self.n_envs = env.n_envs
        self.observation_space = env.observation_space
        self.action_space = env.action_space

    @abc.abstractmethod
    def setup(self):
        """Defines the procedure to setup this model for training"""
        raise NotImplementedError

    @abc.abstractmethod
    def call(self):
        """Forward model"""
        raise NotImplementedError

    @abc.abstractmethod
    def predict(self):
        """Make action predictions for one sample or one batch of samples"""
        raise NotImplementedError

    @abc.abstractmethod
    def _collect_step(self, obs):
        """Collect one step rollout from the environment.

        Args:
            obs (np.ndarray): current observations.

        Returns:
            np.ndarray: next observations
        """
        raise NotImplementedError

    def collect(self, steps, obs=None):
        """Collect `steps` of rollouts. Reset the environments if
        `obs` is not given.

        Args:
            steps (int): number of steps to collect
            obs (np.ndarray, optional): current observations. Defaults to None.
                set to None to reset the envs.

        Returns:
            np.ndarray: next observations
        """
        if obs is None:
            obs = self.env.reset()
        for _ in range(steps):
            obs = self._collect_step(obs)
            # update state
            self.num_timesteps += self.n_envs
        return obs

    def run(self, steps, obs=None):
        """Run the rollout collection procedures can do some
        buffer preprocessing, e.g. compute GAE here.

        Args:
            steps (int): number of steps to collect
            obs (np.ndarray, optional): current observations. Defaults to None.

        Returns:
            np.ndarray: current observations
        """
        return self.collect(steps, obs=obs)

    def get_eval_metrics(self, results):
        """Compute evaluation metrics from evaluation results

        Args:
            results(dict): evaluation results

        Returns:
            dict: evaluation metrics
        """
        if not results:
            return None

        results = ub_utils.flatten_dicts(results)

        rews = results['rewards']
        steps = results['steps']

        metrics = {
            'mean-reward': np.mean(rews).item(),
            'std-reward': np.std(rews).item(),
            'max-reward': np.max(rews).item(),
            'min-reward': np.min(rews).item(),
            'mean-steps': np.mean(steps).item(),
            'std-steps': np.std(steps).item(),
            'max-steps': np.max(steps).item(),
            'min-steps': np.min(steps).item()
        }
        return metrics

    def get_save_metrics(self, metrics):
        """Compute metrics for model selection when saving the model checkpoint.
        Default is the mean of episode rewards.

        Args:
            metrics(dict): evaluation metrics, usually generated by calling
                `self.get_eval_metrics`.

        Returns:
            Any: metrics for selecting the best model.
        """
        if (not isinstance(metrics, dict) or 'mean-reward' not in metrics):
            return None
        return metrics['mean-reward']

    def eval(self, env, n_episodes, max_steps, render=False):
        """Evaluate model for `n_episodes` episodes and limits the maximum steps
        to `max_steps`

        Args:
            env (gym.Env): a gym environment for evaluation.
            n_episodes (int): number of episodes to evaluate.
            max_steps (int): the maximum number of steps. Set to -1 to run forever
                until the environment done.
            render (bool, optional): whether to render the environment screen.
                Defaults to False.

        Returns:
            list: a list of dict containing the evaluation results per episode.
        """
        if n_episodes <= 0:
            return None

        eps_info = []
        for episode in range(n_episodes):
            obs = env.reset()
            total_rews = 0
            total_steps = 0
            # eval one episode
            while total_steps != max_steps:
                # predict action
                acts = self.predict(obs)
                # step environment
                obs, rew, done, info = env.step(acts)
                total_rews += rew
                total_steps += 1
                if render:
                    env.render('human')
                if done:
                    break
            # append eval results
            eps_info.append({'rewards': total_rews, 'steps': total_steps})
        return eps_info

    @abc.abstractmethod
    def _train_step(self, batch_size):
        """Train for one step (one gradient update).

        Args:
            batch_size(int): mini-batch size.

        Returns:
            dict: loss dict
        """
        raise NotImplementedError

    @abc.abstractmethod
    def _update_target(self):
        """Update target networks, this is called for every `target_update`
        (defined in `learn`) steps of gradient updates.
        """
        raise NotImplementedError

    def train(self, batch_size, subepochs, gradsteps, target_update):
        """Train one epoch

        Args:
            batch_size (int): size of minibatch.
            subepochs (int): number of subepochs
            gradsteps (int): Number of steps of gradient updates per subepoch
            target_update (int): target networks update frequency.

        Returns:
            dict: loss dict
        """
        all_losses = []
        for subepoch in range(subepochs):
            for gradstep in range(gradsteps):
                # train one step
                losses = self._train_step(batch_size)
                all_losses.append(losses)
                # update state
                self.num_gradsteps += 1
                # update target networks
                if self.num_gradsteps % target_update == 0:
                    self._update_target()
            # update states
            self.num_subepochs += 1
        # aggregate losses
        all_losses = ub_utils.flatten_dicts(all_losses)
        m_losses = {}
        for name, losses in all_losses.items():
            m_losses[name] = np.mean(np.hstack(np.asarray(losses)))
        return m_losses

    def log_train(
        self,
        total_timesteps: int,
        total_epochs: int,
        time_start: float,
        last_time_spent: float,
        losses: dict,
        verbose: int = 2
    ):
        """Print training log

        Verbosity:
        0 = silent
        1 = silent
        2 = print number of epochs, timesteps, progress... etc
        3 = print full training messages, includes due time, losses...

        Args:
            total_timesteps (int): total training timesteps
            total_epochs (int): total training epochs
            time_start (float): timestamp when the learning started (sec)
            last_time_spent (float): time spent from the learning started
            losses (dict): loss dictionary
            verbose (int, optional): verbosity. Defaults to 2.
        """
        # current time
        time_now = time.time()
        # total time spent
        time_spent = (time_now - time_start)
        # execution time (one epoch)
        execution_time = time_spent - last_time_spent
        # remaining time
        remaining_time = (time_spent / self.progress) * (1.0 - self.progress)
        # eta
        eta = datetime.datetime.now() + datetime.timedelta(seconds=remaining_time)
        # average steps per second
        fps = float(self.num_timesteps) / time_spent

        if verbose > 1:
            self.LOG.set_header(f'Epoch {self.num_epochs}/{total_epochs}')
            self.LOG.add_line()
            self.LOG.add_row(f'Timesteps: {self.num_timesteps}/{total_timesteps}')
            self.LOG.add_row(f'Steps/sec: {fps:.2f}')
            self.LOG.add_row(f'Progress: {self.progress*100.0:.2f}%')
            if verbose > 2:
                self.LOG.add_row(
                    'Execution time', datetime.timedelta(seconds=execution_time)
                )
                self.LOG.add_row(
                    'Elapsed time', datetime.timedelta(seconds=time_spent)
                )
                self.LOG.add_row(
                    'Remaining time', datetime.timedelta(seconds=remaining_time)
                )
                self.LOG.add_row('ETA', eta.strftime('%Y-%m-%d %H:%M:%S'))
                self.LOG.add_line()

                if self.is_warming_up():
                    self.LOG.add_row('Collecting rollouts')
                else:
                    name = '\n'.join(map(str, losses.keys()))
                    loss = '\n'.join(map('{:.6f}'.format, losses.values()))
                    self.LOG.add_rows(
                        fmt='{name} {||} {loss}', name=name, loss=loss
                    )

            self.LOG.add_line()
            self.LOG.flush('INFO')

    def log_eval(self, results, metrics, verbose=2):
        """Print evaluation log

        Verbosity:
        0 = silent
        1 = silent
        2 = print evaluation metrics
        3 = print episode results and metrics

        Args:
            results (list): evaluaition results, returned from `eval`
            metrics (dict): evaluation metrics, returned from `get_eval_metrics`
            verbose (int, optional): verbosity. Defaults to None.
        """
        if verbose > 1:
            if verbose > 2 and results is not None:
                self.LOG.set_header('Evaluation results')
                n_episodes = len(results)
                for ep in range(n_episodes):
                    self.LOG.subgroup(f'Episode {ep+1}')
                    labels = '\n'.join(results[ep].keys())
                    values = '\n'.join(map('{}'.format, results[ep].values()))

                    self.LOG.add_rows(
                        fmt='{labels} {||} {values}', labels=labels, values=values
                    )
                self.LOG.flush('INFO')

            if metrics is not None:
                self.LOG.set_header('Evaluation metrics')
                labels = '\n'.join(metrics.keys())
                values = '\n'.join(map('{:.3f}'.format, metrics.values()))
                self.LOG.add_rows(
                    fmt='{labels} {||} {values}', labels=labels, values=values
                )
                self.LOG.add_line()
                self.LOG.flush('INFO')

    def log_save(self, save_path, saved_path, verbose=2):
        """Print model checkpoint messages

        Verbosity:
        0 = silent
        1 = silent
        2 = print message
        3 = print message

        Args:
            save_path (str): root path to save the checkpoint
            saved_path (str): full path where the checkpoint actually saved
            verbose (int, optional): verbosity. Defaults to None.
        """
        if verbose > 1:
            # find the best model path
            best_path = self._preload(save_path, best=True)
            if best_path == os.path.abspath(saved_path):
                self.LOG.info(f'[Best] Checkpoint saved to: {saved_path}')
            else:
                self.LOG.info(f'Checkpoint saved to: {saved_path}')

    def is_warming_up(self):
        """Determine whether it is in the warmming up stage"""
        return self.num_timesteps < self.warmup_steps

    def learn(
        self,
        total_timesteps: int,
        log_interval: int = 100,
        eval_env: gym.Env = None,
        eval_interval: int = 100,
        eval_episodes: int = 5,
        eval_max_steps: int = 1000,
        save_interval: int = 100,
        save_path: str = None,
        target_update: int = 2,
        tb_logdir: str = None,
        reset_timesteps: bool = False,
        verbose: int = 1
    ):
        """Start the training procedure

        The total number of epochs is calculated by:
        ```
        timesteps_per_epoch = n_steps * n_envs
        total_epochs = total_timesteps / timesteps_per_epoch
        ```

        Args:
            total_timesteps (int): Total number of timesteps to train the model.
            log_interval (int, optional): log frequency (unit: epoch)
                Defaults to 100.
            eval_env (gym.Env, optional): gym environment for evaluation.
                Defaults to None.
            eval_interval (int, optional): evaluation frequency (unit: epoch).
                Defaults to 100.
            eval_episodes (int, optional): evaluation episodes. Defaults to 5.
            eval_max_steps (int, optional): maximum number of steps of each
                episode when evaluating. Defaults to 1000.
            save_interval (int, optional): checkpoint frequency (unit: epoch).
                Defaults to 100.
            save_path (str, optional): root path or directory to save checkpoints.
                Defaults to None.
            target_update (int, optional): frequency of updating the target model.
                (unit: gradstep) Defaults to 2.
            tb_logdir (str, optional): path to store tensorboard files. None to
                disable logging. Defaults to None.
            reset_timesteps (bool, optional): reset the training states.
                Defaults to False.
            verbose (int, optional): more logging. 0=silent, 1=progress bar,
                2=short, 3=full. Defaults to 1.

        Raises:
            RuntimeError: if the training `env` is not set.

        Returns:
            self
        """
        if self.env is None:
            raise RuntimeError('Env not set, call `set_env` before training')

        # create tensorboard writer
        if tb_logdir is not None:
            self.tb_writer = tf.summary.create_file_writer(tb_logdir)

        # initialize state
        if reset_timesteps:
            self.reset()

        # initialze
        obs = None
        time_start = time.time()
        time_spent = 0
        timesteps_per_epoch = self.n_steps * self.n_envs
        total_epochs = math.ceil(
            float(total_timesteps - self.num_timesteps)
            / float(timesteps_per_epoch)
        ) + self.num_epochs
        progress_bar = tqdm.tqdm(total=total_epochs, disable=(verbose != 1))
        progress_bar.update(self.num_epochs)

        while self.num_timesteps < total_timesteps:
            # collect rollouts
            obs = self.run(steps=self.n_steps, obs=obs)

            # update state
            self.num_epochs += 1
            self.progress = float(self.num_timesteps) / float(total_timesteps)
            progress_bar.update()
            # train one epoch
            losses = {}
            if not self.is_warming_up():
                # training
                losses = self.train(
                    batch_size=self.batch_size,
                    subepochs=self.n_subepochs,
                    gradsteps=self.n_gradsteps,
                    target_update=target_update
                )
                # write tensorboard
                if self.tb_writer is not None:
                    with self.tb_writer.as_default():
                        for name, value in losses.items():
                            tf.summary.scalar(
                                f'train/{name}', value, step=self.num_timesteps
                            )
                    self.tb_writer.flush()
            # print training log
            if (log_interval is not None) and (self.num_epochs % log_interval == 0):
                # print log
                self.log_train(
                    total_timesteps,
                    total_epochs,
                    time_start,
                    time_spent,
                    losses,
                    verbose
                )
                # update time_spent
                time_spent = time.time() - time_start

            # evaluate model
            eval_metrics = None
            if (eval_env is not None) and (self.num_epochs % eval_interval == 0):
                eval_results = self.eval(
                    env=eval_env,
                    n_episodes=eval_episodes,
                    max_steps=eval_max_steps
                )

                eval_metrics = self.get_eval_metrics(eval_results)

                if self.tb_writer is not None:
                    with self.tb_writer.as_default():
                        for name, value in eval_metrics.items():
                            tf.summary.scalar(
                                f'eval/{name}', value, step=self.num_timesteps
                            )
                    self.tb_writer.flush()

                self.log_eval(eval_results, eval_metrics, verbose)

            # save model
            if ((save_path is not None) and (save_interval is not None)
                    and (self.num_epochs % save_interval) == 0):
                # compute metrics
                checkpoint_metrics = self.get_save_metrics(eval_metrics)
                saved_path = self.save(
                    save_path,
                    checkpoint_number=self.num_epochs,
                    checkpoint_metrics=checkpoint_metrics
                )
                self.log_save(save_path, saved_path, verbose)
        progress_bar.close()
        return self

    def get_config(self):
        """Get constructor configurations."""
        config = super().get_config()
        config.update({
            'n_steps': self.n_steps,
            'n_subepochs': self.n_subepochs,
            'n_gradsteps': self.n_gradsteps,
            'batch_size': self.batch_size,
            'observation_space': self.observation_space,
            'action_space': self.action_space
        })
        return config

    @classmethod
    def from_config(cls, config):
        """Restore model from the given configurations

        Args:
            config(dict): configurations
        """

        # construct method
        self = cls(env=None, **config)
        # setup model
        self.setup()

        return self

    def save_config(self, filepath):
        """Save model config to `filepath`

        Args:
            filepath(str): path to save configuration
        """
        _config = self.get_config()
        config = {'state': self.state, 'config': _config}

        ub_utils.safe_json_dump(filepath, config)


class OffPolicyModel(BaseRLModel):
    """Defines the default arguments for off-policy models"""
    def __init__(
        self,
        env,
        n_steps: int = 4,
        n_subepochs: int = 1,
        n_gradsteps: int = 1,
        warmup_steps: int = 10000,
        batch_size: int = 128,
        observation_space=None,
        action_space=None,
        **kwargs
    ):
        super().__init__(
            env=env,
            n_steps=n_steps,
            n_subepochs=n_subepochs,
            n_gradsteps=n_gradsteps,
            warmup_steps=warmup_steps,
            batch_size=batch_size,
            observation_space=observation_space,
            action_space=action_space,
            **kwargs
        )


class OnPolicyModel(BaseRLModel):
    """Defines the default arguments for on-policy models.
    Note that in default the target model updates, replay
    buffer warming up stage and `n_gradsteps` is disabled
    for on-policy models. But one can still access them
    by overriding them.
    """
    def __init__(
        self,
        env,
        n_steps: int = 256,
        n_subepochs: int = 4,
        n_gradsteps: int = None,
        warmup_steps: int = None,
        batch_size: int = 128,
        observation_space=None,
        action_space=None,
        **kwargs
    ):
        super().__init__(
            env=env,
            n_steps=n_steps,
            n_subepochs=n_subepochs,
            n_gradsteps=n_gradsteps,
            warmup_steps=warmup_steps,
            batch_size=batch_size,
            observation_space=observation_space,
            action_space=action_space,
            **kwargs
        )

    def _update_target(self):
        """Do nothing"""
        pass

    def is_warming_up(self):
        """On policy algo has no wram up stage"""
        return False



import os
import sys
import time
import string
import logging


MODULE_PATH = __name__.rsplit('.', 1)[0]
c_module = __import__(MODULE_PATH+'.cmd_colors', fromlist=['.'])
C = getattr(c_module, 'Color')


font_module = __import__(MODULE_PATH+'.ascii_font.__init__', fromlist=['.'])
ascii_font = font_module #getattr(font_module, 'ascii_font')

#from sshex.logger.cmd_colors import Color as C


def format_color_message(message, code_map, color_fmt):
    for code in code_map:
        message = message.replace(code, color_fmt.format(code_map[code]))
    return message


class ColoredFormatter(logging.Formatter):
 
    LEVEL_COLORS = {
        'WARNING': C.BOLD + C.dark(C.YELLOW),
        'INFO': C.BOLD + C.GREEN,
        'ERROR': C.BOLD + C.dark(C.RED),
        'DEBUG': C.BOLD + C.dark(C.BLUE),
        'CRITICAL': C.BOLD + C.RED
    }

    LEVEL_LABEL = {
        'WARNING': 'WARN',
        'INFO': 'INFO',
        'ERROR': 'ERRO',
        'DEBUG': 'DEBU',
        'CRITICAL': 'ERRO'
    }

    CODES = {
        '$TIME': C.dark(C.GREEN),
        '$THREAD': C.light(C.BLACK),
        '$MODULE': C.dark(C.CYAN),
        '$BOLD': C.BOLD,
        '$UNBOLD': C.reset(C.BOLD),
        '$BLINK': C.BLINK,
        '$RESET': C.reset(),
    }

    COLOR_FMT = '\033[{}m'

    def __init__(self, fmt=None, datefmt=None, style='%'):

        if fmt is not None:
            fmt = format_color_message(fmt, self.CODES, self.COLOR_FMT)

        super(ColoredFormatter, self).__init__(fmt, datefmt, style)


    def getLevelColor(self, record):
        return self.LEVEL_COLORS[record.levelname]


    def format(self, record):
        # get $LEVEL color
        levelcolor = self.getLevelColor(record)

        

        # set level label
        original_levelname = record.levelname
        record.levelname = self.LEVEL_LABEL[record.levelname]

    
        # === formatting ===
        record.message = record.getMessage()
        # asctime
        if self.usesTime():
            record.asctime = self.formatTime(record, self.datefmt)
        
        # message
        s = self.formatMessage(record)

        # exception
        if record.exc_info:
            if not record.exc_text:
                record.exc_text = self.formatException(record.exc_info)

        if record.exc_text:
            if s[-1:] != '\n':
                s = s + '\n'
            s = s + '$EXCEPT' + record.exc_text + '$RESET'
        if record.stack_info:
            if s[-1:] != '\n':
                s = s + '\n'
            s = s + '$EXCEPT' + self.formatStack(record.stack_info) + '$RESET'

        self.CODES['$EXCEPT'] = self.LEVEL_COLORS['ERROR']
        self.CODES['$LEVEL'] = levelcolor

        s = format_color_message(s, self.CODES, self.COLOR_FMT)


        record.levelname = original_levelname

        return s


def _recursive_format(value, args):
    '''
    :returns: formatted single str object
    '''

    if len(args) == 0:
        return format(value, '') # args = []
    elif len(args) == 1:
        return format(value, args[0]) # args ['03d']

    # --- magic font ---
    elif '@f' in args:
        #print(args)
        index = args.index('@f')
        #print(index)
        if len(args) <= index+1:
            raise ValueError('No font type specified.')

        font_type = args[index+1]

        del args[index+1]
        del args[index]
    
        # must be a str object
        # !! recursive call !!
        formatted_obj = _recursive_format(value, args)

        # get ascii font for each character
        objs = [obj.split('\n') for obj in ascii_font.retrieves(font_type, formatted_obj)]
        
        # concat characters into a list of lines
        lines = multilineformat(objs)

        return '\n'.join(lines) #join to a single str

    # error
    return None


def customformat(value, format_spec=''):

    # split by ':'
    # :@f:font:03d
    args = format_spec.split(':')


    formatted_obj = _recursive_format(value, args)


    if formatted_obj is None:
        raise ValueError('Unknown format code \'{}\' for object of type \'{}\''.format(format_spec, type(value)))

    return formatted_obj.split('\n')


def calculatemaxline(value_list):
    max_count = 0
    for v in value_list:
        if isinstance(v, list) or isinstance(v, tuple):
            max_count = max(max_count, len(v))

    return max_count


def multilineformat(result, max_line_count=None):


    if max_line_count is None:
        max_line_count = calculatemaxline(result)


    # multi line formatting
    lines = ['' for _ in range(max_line_count)]

    current_max_length = 0

    for idx, segment in enumerate(result):

        # literal_text
        if isinstance(segment, str):
            # only the first line will add literal_text
            lines[0] += segment
            # others will append spaces
            for line_id in range(1, max_line_count):
                lines[line_id] += ' ' * len(segment)
                    
            current_max_length += len(segment)

        # formatting
        elif isinstance(segment, tuple):
            spec = segment[0]
            pattern = segment[1]

            if spec == 'VERT':
                for line_id in range(max_line_count):
                    lines[line_id] += pattern

                current_max_length += len(pattern)

        # formatted object
        else:
            for line_id in range(len(segment)):
                lines[line_id] += segment[line_id]

                # update max length
                current_max_length = max(current_max_length, len(lines[line_id]))
            

        # padding each line to current_max_length
        for line_id in range(max_line_count):
            lines[line_id] += ' ' * (current_max_length - len(lines[line_id])) # append spaces



    return lines

class MultilineStringFormatter(string.Formatter): 

    def format_field(self, value, format_spec):
        return customformat(value, format_spec)

    def _vformat(self, format_string, args, kwargs, used_args, recursion_depth, auto_arg_index=0):

        if recursion_depth < 0:
            raise ValueError('Max string recursion exceeded')

        result = []

        max_line_count = 0

        for literal_text, field_name, format_spec, conversion in self.parse(format_string):

            if literal_text:
                result.append(literal_text)

            if field_name is not None:

                # auto field
                if field_name == '':
                    if auto_arg_index is False:
                        raise ValueError('cannot switch from manual field specification to automatic field numbering')

                    field_name = str(auto_arg_index)
                    auto_arg_index += 1

                elif field_name.startswith('|'):
                    # add vertical line
                    if field_name[1:] == '':
                        field_name = '|'
                    else:
                        field_name = field_name[1:]

                    result.append(('VERT', field_name))
                    
                    continue

                elif field_name.isdigit():
                    if auto_arg_index:
                        raise ValueError('cannot switch from manual field specification to automatic field numbering')

                    auto_arg_index = False

                # given the field_name, find the object it references
                obj, arg_used = self.get_field(field_name, args, kwargs)
                used_args.add(arg_used)

                obj = self.convert_field(obj, conversion)

                format_spec, auto_arg_index = string.Formatter()._vformat(
                            format_spec, args, kwargs,
                            used_args, recursion_depth-1,
                            auto_arg_index=auto_arg_index)

                # splitted by \n
                formatted_obj = self.format_field(obj, format_spec)


                # format the object and append to the result
                result.append(formatted_obj)

                max_line_count = max(max_line_count, len(formatted_obj))

        lines = multilineformat(result, max_line_count)
        
        '''
        # multi line formatting
        lines = ['' for _ in range(max_line_count)]

        current_max_length = 0

        for idx, segment in enumerate(result):

            # literal_text
            if isinstance(segment, str):
                # only the first line will add literal_text
                lines[0] += segment
                # others will append spaces
                for line_id in range(1, max_line_count):
                    lines[line_id] += ' ' * len(segment)
                    
                current_max_length += len(segment)

            # formatting
            elif isinstance(segment, tuple):
                spec = segment[0]
                pattern = segment[1]

                if spec == 'VERT':
                    for line_id in range(max_line_count):
                        lines[line_id] += pattern

                    current_max_length += len(pattern)

            # formatted object
            else:
                for line_id in range(len(segment)):
                    lines[line_id] += segment[line_id]

                    # update max length
                    current_max_length = max(current_max_length, len(lines[line_id]))
            

            # padding each line to current_max_length
            for line_id in range(max_line_count):
                lines[line_id] += ' ' * (current_max_length - len(lines[line_id])) # append spaces
        '''

        return lines, auto_arg_index




# --- built in ---
import os
import sys
import time
import random
import datetime

from distutils.version import StrictVersion as Version

# --- 3rd party ---
import gym
import cloudpickle

import numpy as np
import tensorflow as tf

# --- my module ---

__all__ = [
    'ReLU'
]

# --- Some TensorFlow internal bugs fixes ---

# TODO: TensorFlow 2.4 Internal bugs (or caused by CUDA 11?)
# tf.keras.layers.ReLU will cause memory leaking, so use tf.maximum instead
# refer to this github issue: https://github.com/tensorflow/tensorflow/issues/46475
if Version(tf.__version__) > '2.4.0':
    class ReLU(tf.keras.layers.Layer):
        def __init__(self, **kwargs):
            super().__init__(**kwargs)

        def build(self, input_shape):
            pass
    
        def call(self, inputs, training=False):
            return tf.math.maximum(inputs, 0.)
else:
    ReLU = tf.keras.layers.ReLU




import ascii_font

ascii_font.load_font_type_list()

font_types = list(ascii_font.FONT_TYPE_LIST.keys())

print(font_types)

for font_type in font_types:

    print('loading font type: {}'.format(font_type))
    font = ascii_font.load(font_type)

    for c in ['a', 'b', 'c', ' ', '!']:
        
        print(font.retrieve(c))



import cloudpickle
from .lib import base
from .lib import data
from .lib import envs
from .lib import nets
from .lib import prob
from .lib import sche
from .lib import patch
from .lib import utils

__copyright__ = '''
 The MIT License (MIT)
 Copyright (c) 2021 Joe Hsiao
 
 Permission is hereby granted, free of charge, to any person obtaining a copy
 of this software and associated documentation files (the "Software"), to deal
 in the Software without restriction, including without limitation the rights
 to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
 copies of the Software, and to permit persons to whom the Software is
 furnished to do so, subject to the following conditions:
 
 The above copyright notice and this permission notice shall be included in all
 copies or substantial portions of the Software.
 
 THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
 EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
 MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
 IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM,
 DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR
 OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE
 OR OTHER DEALINGS IN THE SOFTWARE.
'''
__license__ = 'MIT'
__version__ = '0.1.0a0'



# --- built in ---
import os
import re

from setuptools import find_packages, setup

def get_version():
    with open(os.path.join('unstable_baselines', '__init__.py'), 'r') as f:
        return re.search(r'^__version__ = [\'"]([^\'"]*)[\'"]', f.read(), re.M).group(1)

setup(
    name='unstable_baselines',
    version=get_version(),
    description='A TensorFlow 2.0 implemented RL baselines',
    long_description=open('README.md', encoding='utf8').read(),
    long_description_content_type='text/markdown',
    url='https://github.com/Ending2015a/unstable_baselines',
    author='JoeHsiao',
    author_email='joehsiao@gapp.nthu.edu.tw',
    license='MIT',
    python_requires=">=3.6",
    classifiers=[
        # How mature is this project? Common values are
        #   3 - Alpha
        #   4 - Beta
        #   5 - Production/Stable
        'Development Status :: 2 - Pre-Alpha',
        # Indicate who your project is intended for
        'Intended Audience :: Science/Research',
        'Topic :: Scientific/Engineering :: Artificial Intelligence',
        'Topic :: Software Development :: Libraries :: Python Modules',
        # Pick your license as you wish (should match "license" above)
        'License :: OSI Approved :: MIT License',
        # Specify the Python versions you support here. In particular, ensure
        # that you indicate whether you support Python 2, Python 3 or both.
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.8',
        'Programming Language :: Python :: 3.9',
    ],
    keywords='reinforcement-learning playform tensorflow-2.0 tensorflow',
    packages=[
        # exclude deprecated module
        package for package in find_packages(exclude=["*.dep.*", "*.dep"])
        if package.startswith('unstable_baselines')
    ],
    package_data={
        'unstable_baselines.logger': ['ascii_font/pkl/*.pkl']
    },
    install_requires=[
        'gym==0.18.0',
        'tensorflow>=2.3.0',
        'numpy',
        'cloudpickle',
        'opencv-python',
        'tqdm'
    ],
    extras_require={
        'dev': [
            # Unittest and coverage
            'coverage',
            'parameterized',
            'scipy',
            'yapf',
            'flake8',
            'flake8-bugbear',
        ],
        'atari': ['atari_py==0.2.6'],
        'pybullet': ['pybullet'],
        'extra': [
            'atari_py==0.2.6',
            'pybullet',
            'ray[tune]'
        ]
    }
)



# --- built in ---
import os
import abc
import sys

from typing import Union

# --- 3rd party ---
import gym
import numpy as np

# --- my module ---
from unstable_baselines.lib import utils as ub_utils

[lambda i=j:j for j in range(10)]

# The implementation mainly follows tianshou/venvs.py
# habitat-lab/vector_env.py and stable-baselines3/subproc_vec_env.py
class BaseEnvWorker(metaclass=abc.ABCMeta):
    def __init__(self, env_fn, auto_reset: bool):
        self._env_fn = env_fn
        self._auto_reset = auto_reset

    def setup(self):
        self.observation_space = self.getattr('observation_space')
        self.action_space      = self.getattr('action_space')
        self.metadata          = self.getattr('metadata')
        self.reward_range      = self.getattr('reward_range')
        self.spec              = self.getattr('spec')

    @abc.abstractmethod
    def getattr(self, attrname: str):
        raise NotImplementedError

    @abc.abstractmethod
    def setattr(self, attrname: str, value):
        raise NotImplementedError

    @abc.abstractmethod
    def reset(self, **kwargs):
        raise NotImplementedError

    @abc.abstractmethod
    def step_async(self, act):
        raise NotImplementedError

    @abc.abstractmethod
    def step_wait(self):
        raise NotImplementedError

    @abc.abstractmethod
    def seed(self, seed):
        self.action_space.seed(seed)

    @abc.abstractmethod
    def render(self):
        raise NotImplementedError

    @abc.abstractmethod
    def close_async(self):
        raise NotImplementedError
    
    @abc.abstractmethod
    def close_wait(self):
        raise NotImplementedError

# The implementation mainly follows tianshou/venvs.py
# habitat-lab/vector_env.py and stable-baselines3/subproc_vec_env.py
class BaseVecEnv(gym.Env):
    def __init__(self,
        env_fns:      list,
        worker_class: BaseEnvWorker,
        rms_norm:     Union[str, bool, ub_utils.RMSNormalizer] = False,
        auto_reset:   bool = True
    ):
        self._worker_class = worker_class
        self.workers       = [worker_class(fn, auto_reset) for fn in env_fns]
        self.n_envs        = len(self.workers)
        self._closed       = False
        # setup workers
        [w.setup() for w in self.workers]

        # init properties
        self.observation_spaces = [w.observation_space for w in self.workers]
        self.action_spaces      = [w.action_space for w in self.workers]
        self.metadatas          = [w.metadata for w in self.workers]
        self.reward_ranges      = [w.reward_range for w in self.workers]
        self.specs              = [w.spec for w in self.workers]

        self.observation_space = self.observation_spaces[0]
        self.action_space      = self.action_spaces[0]
        self.metadata          = self.metadatas[0]
        self.reward_range      = self.reward_ranges[0]
        self.spec              = self.specs[0]

        self._rms_norm = self._get_rms_norm_opt(rms_norm)

    @property
    def rms_norm(self):
        return self._rms_norm

    @property
    def closed(self):
        return self._closed

    def __len__(self):
        return self.n_envs

    def getattrs(self, attrname: str, id: list=None):
        '''Retrieve attributes from each environment

        Args:
            attrname (str): Attribute name
            id (list, optional): Environment indices. Defaults to None.

        Returns:
            list: A list of retrieved attributes
        '''
        self._assert_not_closed()
        ids = self._get_ids(id)
        return [self.workers[id].getattr(attrname) for id in ids]

    def setattrs(self, attrname:str, value=None, values: list=None, id: list=None):
        '''Set attributes for each environments

        Args:
            attrname (str): Attribute name
            value (Any, optional): Attribute value. This value will be passed
                to all envs specified by `id`. Defaults to None.
            values (list, optional): A list of attribute values. Defaults to None.
            id (list, optional): Environment indices. Defaults to None.
        
        Returns:
            list: Returned values
        '''
        self._assert_not_closed()
        if values is not None and value is not None:
            raise ValueError('Both `value` and `values` are set. '
                'Only one of them can be specified.')
        ids = self._get_ids(id)
        if values is not None:
            assert len(values) == len(ids)
            return [self.workers[id].setattr(attrname, v) 
                    for id, v in zip(ids, values)]
        return [self.workers[id].setattr(attrname, value) for id in ids]

    def reset(self, id: list=None, **kwargs):
        self._assert_not_closed()
        ids = self._get_ids(id)
        obs_list = [self.workers[id].reset(**kwargs) for id in ids]
        obs = self._flatten_obs(obs_list)
        self.rms_norm.update(obs)
        return self.rms_norm(obs)

    def step(self, acts: np.ndarray, id:list=None):
        self._assert_not_closed()
        ids = self._get_ids(id)
        assert len(acts) == len(ids)
        # step workers
        [self.workers[id].step_async(act) for act, id in zip(acts, ids)]
        results = [self.workers[id].step_wait() for id in ids]
        obs_list,rew_list,done_list,info_list = zip(*results)
        # stack results
        obs  = self._flatten_obs(obs_list)
        rew  = np.stack(rew_list)
        done = np.stack(done_list)
        # rms norm
        self.rms_norm.update(obs)
        return self.rms_norm(obs), rew, done, info_list

    def seed(self, seed:list=None, id:list=None):
        self._assert_not_closed()
        ids = self._get_ids(id)
        if seed is None:
            seed_list = [None] * self.n_envs
        elif isinstance(seed, int):
            seed_list = [seed+id for id in ids]
        else:
            # list, tuple, np.ndarray
            assert len(seed) == len(ids)
            seed_list = seed
        return [self.workers[id].seed(s) for id, s in zip(ids, seed_list)]

    def render(self, **kwargs):
        self._assert_not_closed()
        return [w.render(**kwargs) for w in self.workers]

    def close(self):
        if self.closed:
            return
        [w.close_async() for w in self.workers]
        [w.close_wait()  for w in self.workers]
        self._closed = True

    def _flatten_obs(self, obs_list):
        obs = ub_utils.map_nested_tuple(
            tuple(obs_list), lambda obs: np.stack(obs))
        return obs

    def _get_ids(self, id: list=None):
        if id is None:
            return list(range(self.n_envs))
        return [id] if np.isscalar(id) else id

    def _assert_not_closed(self):
        assert not self.closed, "This env is already closed"

    def _get_rms_norm_opt(self, opt):
        if isinstance(opt, ub_utils.RMSNormalizer):
            # setup RMS normalizer
            if not opt.is_setup:
                opt.setup(self.observation_space)
            return opt
        # select default RMS normalizer
        if isinstance(self.observation_space, 
                (gym.spaces.Dict, gym.spaces.Tuple)):
            _rms_class = ub_utils.NestedRMSNormalizer
        else:
            _rms_class = ub_utils.RMSNormalizer
        enable = None
        if isinstance(opt, bool):
            enable = opt
        # Create RMSNormalizer
        rms = _rms_class(space  = self.observation_space, 
                         enable = enable)
        if isinstance(opt, str):
            rms.load(opt)
        return rms


# --- bulid in ---
import os
import abc
import csv
import json
import time
import tempfile
import distutils
import subprocess

# --- 3rd party ---
import gym
import numpy as np

# --- my module ---
from unstable_baselines import logger
from unstable_baselines.lib import utils as ub_utils

LOG = logger.getLogger()

__all__ = [
    'Monitor',
    'MonitorToolChain',
    'StatsRecorder',
    'VideoRecorder',
]

# === Monitor ===

class Monitor(gym.Wrapper):
    def __init__(self, env:            gym.Env,
                       root_dir:           str = './monitor',
                       prefix:             str = None,
                       allow_early_reset: bool = True,
                       video:             bool = False,
                       video_kwargs:      dict = {}):
        '''A monitor wrapper to record environment episode reward, length, time
        and game screen.

        Args:
            env (gym.Env): Environment.
            root_dir (str, optional): Root path to store monitor files. Defaults 
                to './monitor'.
            prefix (str, optional): Monitor file prefix. Defaults to None.
            allow_early_reset (bool, optional): If False, monitor ignores the
                episodes that are reset before done. Defaults to True.
            video (bool, optional): Whether to record video (mp4). This feature 
                only works for the environments which support 'rgb_array' render 
                mode. Defaults to False.
            video_kwargs (dict, optional): Recording keywarded arguments. See
                VideoRecorder for more info. Defaults to {}.
        '''        
        super().__init__(env)

        self.root_dir = root_dir or './'
        self.prefix   = prefix
        self.env_id   = env.unwrapped.spec.id
        self.allow_early_reset = allow_early_reset

        self._stats_recorder = StatsRecorder(
            root_dir          = self.root_dir,
            prefix            = self.prefix,
            ext               = None,
            allow_early_reset = self.allow_early_reset
        )
        self._tools = []

        self.add_tool(self._stats_recorder)
        if video:
            video_kwargs = video_kwargs or {}
            self.add_tool(VideoRecorder(**video_kwargs))
    
    @property
    def stats(self):
        '''Return StateRecorder'''
        return self._stats_recorder._stats

    @property
    def tools(self):
        '''Return Toolchain'''
        return self._tools

    def add_tool(self, tool: 'MonitorToolChain'):
        '''Add monitor tool'''
        if hasattr(tool, 'set_monitor'):
            tool.set_monitor(self)
        self._tools.append(tool)
        return self

    def add_tools(self, tool_list: list):
        '''Add monitor toolchain'''
        for tool in tool_list:
            self.add_tool(tool)
        return self

    def step(self, act):
        act = self._before_step(act)
        obs, rew, done, info = self.env.step(act)
        return self._after_step(obs, rew, done, info)

    def reset(self, **kwargs):
        kwargs = self._before_reset(**kwargs)
        obs = self.env.reset(**kwargs)
        return self._after_reset(obs)

    def close(self):
        for tool in self._tools:
            tool.close()
        super().close()
    
    def _before_step(self, act):
        for tool in self._tools:
            act = tool._before_step(act)
        return act
    
    def _after_step(self, *data):
        for tool in self._tools:
            data = tool._after_step(*data)
        return data

    def _before_reset(self, **kwargs):
        for tool in self._tools:
            kwargs = tool._before_reset(**kwargs)
        return kwargs

    def _after_reset(self, obs):
        for tool in self._tools:
            obs = tool._after_reset(obs)
        return obs

class MonitorToolChain(metaclass=abc.ABCMeta):
    def __init__(self):
        '''A base class for implementing custom monitor tool chains
        '''
        self._monitor = None
        self._stats   = None
        self._env     = None

    def set_monitor(self, monitor: 'Monitor'):
        self._monitor = monitor
        self._stats   = monitor.stats
        self._env     = monitor.env

    def _before_step(self, act):
        return act

    def _after_step(self, *data):
        return data

    def _before_reset(self, **kwargs):
        return kwargs

    def _after_reset(self, obs):
        return obs

    def close(self):
        pass

# === Stats Recorder ===

class Stats(ub_utils.StateObject):
    def __init__(self):
        super().__init__()
        self.episodes    = 0  # Cirrent episode number
        self.steps       = 0  # Current step number
        self.start_steps = 0  # Step when this episode begins
        self.rewards     = 0  # Cumulative rewards in one episode
        self.ep_rewards  = [] # Rewards per episode
        self.ep_lengths  = [] # Episode lengths per episode
        self.ep_times    = [] # Episode time per episode

class StatsRecorder(MonitorToolChain):
    monitor_ext = 'monitor.csv'
    def __init__(self, root_dir:           str = None, 
                       prefix:             str = None, 
                       ext:                str = None,
                       allow_early_reset: bool = True):
        '''StatsRecorder records environment stats: episodic rewards,
        total timesteps, time spent. And write them to a monitor file
        `{root_dir}/{prefix}.{ext}` in CSV format.

        CSV contents:
        The first line is a JSON comment indicating the timestamp 
        the environment started, and the name of the environment.
        The next line is the header line, indicating the column
        names [r, l, t] for each columns, where 'r' denotes
        rewards, 'l' episode length, 't' episode running time.
        The following lines are data for each episode.

        Example:
            # {t_start: timestamp, env_id: env_id}
            r, l, t
            320, 1000, 5.199693
            ...

        r: total episodic rewards
        l: episode length
        t: running time (time - t_start)

        Args:
            root_dir (str, optional): Root path to save stats files. Defaults 
                to None.
            prefix (str, optional): File prefix. Defaults to None.
            ext (str, optional): File extention. Defaults to 'monitor.csv'.
            env_id (str, optional): Environment ID. Defaults to None.
        '''
        self.t_start = time.time()

        self.root_dir = root_dir
        self.prefix   = prefix
        self.ext      = ext
        self.env_id   = None
        self.allow_early_reset = allow_early_reset

        # initialize
        self._closed      = False
        self._need_reset  = True
        self.filepath     = None
        self.header       = None
        self.file_handler = None
        self.writer       = None

        self._stats = Stats()

    def set_monitor(self, monitor: Monitor):
        self._monitor = monitor
        self.root_dir = self.root_dir or monitor.root_dor
        self.prefix   = self.prefix or monitor.prefix
        self.ext      = self.ext or self.monitor_ext
        self.env_id   = self.env_id or monitor.env_id
        self._setup_writer()

    def _setup_writer(self):
        # setup csv file
        self.filepath = self._make_save_path(self.root_dir, self.prefix, self.ext)
        self.header = json.dumps({'t_start': self.t_start, 'env_id': self.env_id})

        filedir = os.path.dirname(self.filepath)
        relpath = os.path.relpath(self.filepath)

        self._create_path(filedir)
        LOG.debug('Writing monitor to: ' + relpath)

        # write header
        self.file_handler = open(self.filepath, 'wt')
        self.file_handler.write(f'#{self.header}\n')

        # create csv writer
        self.writer = csv.DictWriter(self.file_handler, 
                    fieldnames=('rewards', 'length', 'walltime'))
        self.writer.writeheader()
        self.file_handler.flush()

    @property
    def stats(self):
        return self._stats

    @property
    def closed(self):
        return self._closed

    def _make_save_path(self, base_path, prefix, ext):
        '''Make save path {base_path}/{prefix}.{ext}
        or {base_path}/{ext} if {preffix} is None
        '''
        paths = []
        if prefix:
            paths.append(str(prefix))
        paths.append(str(ext))

        filename = '.'.join(paths)
        path = os.path.join(base_path, filename)

        return os.path.abspath(path)

    def _create_path(self, path):
        '''Ensure nested directories exist
        if not, create them
        '''
        os.makedirs(path, exist_ok=True)

    def _before_step(self, act):
        if self._need_reset:
            raise RuntimeError('Tried to step environment that needs reset')
        return act

    def _after_step(self, obs, rew, done, info):
        # append rewards
        self.stats.rewards += rew
        self.stats.steps += 1
        # episode ended
        if done:
            self._need_reset = True
            ep_info = self._write_results()
            info['episode'] = ep_info
        return obs, rew, done, info

    def _before_reset(self, **kwargs):
        '''Do nothing'''
        return kwargs

    def _after_reset(self, obs):
        '''Reset episodic info'''
        if not self._need_reset:
            if self.allow_early_reset:
                self._write_results()
        self.stats.rewards = 0
        self.stats.start_steps = self.stats.steps
        self.stats.episodes += 1
        self._need_reset = False
        return obs

    def _write_results(self):
        self._need_reset = True
        ep_rew = self.stats.rewards
        ep_len = self.stats.steps - self.stats.start_steps
        ep_time = time.time()-self.t_start
        ep_info = {'rewards': round(ep_rew, 6), 
                    'length': ep_len,
                    'walltime': round(ep_time, 6)}
        self.stats.ep_rewards.append(ep_rew)
        self.stats.ep_lengths.append(ep_len)
        self.stats.ep_times.append(ep_time)
        # write episode info to csv
        if self.writer:
            self.writer.writerow(ep_info)
            self.file_handler.flush()
        return ep_info

    def close(self):
        if self.file_handler is not None:
            self.file_handler.close()
        self._closed = True

    def __del__(self):
        if not self._closed:
            print('WARNING: The StatsRecorder is not closed, please '
                'ensure you have closed the environment before the '
                'program ended `env.close()`')
            self.close()

    def flush(self):
        pass

# === Video Recorder ===

class _VideoEncoder(object):
    def __init__(self, path, 
                       frame_shape, 
                       video_shape, 
                       in_fps, 
                       out_fps):
        '''
        path: output path
        frame_shape: input frame shape (h,w,c)
        video_shape: output video shape (h,w,c)
        in_fps: input framerate
        out_fps: output video framerate
        '''
        self.path = path
        self.frame_shape = frame_shape
        self.video_shape = video_shape
        self.in_fps = in_fps
        self.out_fps = out_fps

        self.proc = None
        h,w,c = self.frame_shape
        oh,ow,_ = self.video_shape
        
        if c != 3 and c != 4:
            raise RuntimeError(f'Your frame has shape {self.frame_shape}, but we require (h,w,3) or (h,w,4)')
        
        self.wh = (w,h)
        self.output_wh = (ow,oh)
        self.includes_alpha = (c == 4)

        if distutils.spawn.find_executable('ffmpeg') is not None:
            self.backend = 'ffmpeg'
        else:
            raise RuntimeError('ffmpeg not found')
        
        self._start_encoder()

    def _start_encoder(self):
        self.cmdline = (
            self.backend,
            '-nostats',   # disable encoding progress info
            '-loglevel', 'error', # suppress warnings
            '-y',  # replace existing file

            # input
            '-f', 'rawvideo', # input format
            '-s:v', "{:d}x{:d}".format(*self.wh), # input shape (w,h)
            '-pix_fmt', ('rgb32' if self.includes_alpha else 'rgb24'),
            '-framerate', '{:d}'.format(self.in_fps), # input framerate,
            '-i', '-', # input from stdin

            # output
            '-vf', 'scale={:d}:{:d}'.format(*self.output_wh),
            '-c:v', 'libx264', # video codec
            '-pix_fmt', 'yuv420p',
            '-r', '{:d}'.format(self.out_fps),
            self.path)
        
        cmd = ' '.join(self.cmdline)
        LOG.debug(f'Starting ffmpeg with cmd "{cmd}"')

        self.proc = subprocess.Popen(self.cmdline, stdin=subprocess.PIPE)
    
    @property
    def version_info(self):
        return {
            'backend': self.backend,
            'version': str(subprocess.check_output([self.backend, '-version'],stderr=subprocess.STDOUT)),
            'cmdline': self.cmdline
        }

    @property
    def video_info(self):
        return {
            'width': self.wh[0],
            'height': self.wh[1],
            'input_fps': self.in_fps,
            'output_fps': self.out_fps
        }

    def capture_frame(self, frame):
        if not isinstance(frame, (np.ndarray, np.generic)):
            raise RuntimeError(f'Wrong type {type(frame)} for {frame} '
                '(must be np.ndarray or np.generic)')
        if frame.shape != self.frame_shape:
            raise RuntimeError(f'Your frame has shape {frame.shape}, '
                'but the VideoRecorder is configured for shape '
                f'{self.frame_shape}')
        if frame.dtype != np.uint8:
            raise RuntimeError(f'Your frame has data type {frame.dtype}'
                ', but we require unit8 (i.e. RGB values from 0-255)')

        self.proc.stdin.write(frame.tobytes())

    def __del__(self):
        if self.proc is not None:
            print('WARNING: The _VideoEncoder is not closed, please '
                    'ensure you have closed the environment before the '
                    'program ended `env.close()`')
            self.close()

    def close(self):
        if self.proc is None:
            return
        
        self.proc.stdin.close()
        ret = self.proc.wait()
        if ret != 0:
            LOG.error(f'VideoRecorder encoder exited with status {ret}')

        self.proc = None


class _VideoRecorder(object):
    def __init__(self, env:   gym.Env, 
                       path:      str, 
                       meta_path: str, 
                       metadata: dict = None, 
                       width:     int = None, 
                       height:    int = None,
                       in_fps:    int = None,
                       out_fps:   int = None,
                       enabled:  bool = True):
        '''VideoRecorder captures video frames
        from `env` and saves to `path`.

        Args:
            env (gym.Emv): Environment.
            path (str): Video save path.
            meta_path (str): Metadata save path.
            metadata (dict, optional): Additional metadata. Defaults to None.
            width (int, optional): Video width. Defaults to None.
            height (int, optional): Video height. Defaults to None.
            in_fps (int, optional): Environment fps. Defaults to None.
            out_fps (int, optional): Output video fps. Defaults to None.
            enabled (bool, optional): Enable this recorder. Defaults to True.
        '''
        self.env       = env
        self.path      = path
        self.meta_path = meta_path
        self.width     = width
        self.height    = height
        self.in_fps    = in_fps
        self.out_fps   = out_fps
        self.enabled   = enabled
        self.metadata  = metadata or {}
        if not self.enabled:
            return

        if not path:
            raise ValueError(f"'path' not specified: {path}")
        if not meta_path:
            raise ValueError(f"'meta_path' not specified: {meta_path}")

        in_fps  = env.metadata.get('video.frames_per_second', 30)
        out_fps = env.metadata.get('video.output_frames_per_second', in_fps)

        self.in_fps  = self.in_fps or self.out_fps or in_fps
        self.out_fps = self.out_fps or self.in_fps or out_fps

        self._async = env.metadata.get('semantics.async')
        modes       = env.metadata.get('render.modes', [])

        if 'rgb_array' not in modes:
            LOG.warning(f'Disabling video recorder because {env} not '
                    'supports render mode "rgb_array"')
            self.enabled = False
            return
        
        # pybullet_envs: set render width, height to the environment
        if (width is not None) and hasattr(env.unwrapped, '_render_width'):
            env.unwrapped._render_width = width
        if (height is not None) and hasattr(env.unwrapped, '_render_height'):
            env.unwrapped._render_height = height

        # Create directory
        self._create_path(os.path.abspath(os.path.dirname(self.path)))
        self._create_path(os.path.abspath(os.path.dirname(self.meta_path)))
        
        # Write metadata
        self.metadata['content_type'] = 'video/mp4'
        self.write_metadata()

        self.encoder = None
        self.broken  = False
        self.empty   = True
        
    @property
    def closed(self):
        '''Return True, if disabled'''
        return not self.enabled

    @property
    def functional(self):
        return self.enabled and not self.broken

    def capture_frame(self):
        if not self.functional:
            return 
        # capture current frame
        frame = self.env.render(mode='rgb_array')
        if frame is None:
            if self._async:
                return
            else:
                LOG.error('Env returned None on render(). Disabling further '
                        'rendering for video recorder by marking as disabled: '
                        f'{self.path}')
                self.broken = True
        else:
            self._encode_image_frame(frame)

    def close(self):
        if not self.enabled:
            return

        if self.encoder:
            LOG.debug(f'Closing video encoder: {self.path}')
            self.encoder.close()
            self.encoder = None
        else:
            if os.path.isfile(self.path):
                os.remove(self.path)
            
            self.metadata['empty'] = True

        if self.broken:
            LOG.warning('Cleaning up paths for broken video recorder: '
                        f'{self.path}')

            if os.path.isfile(self.path):
                os.remove(self.path)

            self.metadata['broken'] = True
        self.write_metadata()
        # mark as closed
        self.enabled = False

    def write_metadata(self):
        if not self.enabled:
            return
        try:
            with open(self.meta_path, 'wt') as f:
                json.dump(self.metadata, f, indent=2)
        except FileNotFoundError:
            LOG.error(f'Failed to write metadata to path: {self.meta_path}\n'
                'This error may caused by trying to write files in a __del__'
                ' function')

    def _create_path(self, path):
        '''Ensure nested directories exist
        if not, create them
        '''
        os.makedirs(path, exist_ok=True)

    def _get_frame_shape(self, frame):
        if len(frame.shape) != 3:
            raise RuntimeError('Receive unknown frame shape: '
                                f'{frame.shape}')

        return frame.shape # h,w,c

    def _get_video_shape(self, frame):
        if len(frame.shape) != 3:
            raise RuntimeError('Receive unknown frame shape: '
                                f'{frame.shape}')

        h,w,c = frame.shape

        self.width = self.width if self.width else w
        self.height = self.height if self.height else h

        return (self.height, self.width, c) # h,w,c

    def _encode_image_frame(self, frame):
        if not self.encoder:
            frame_shape = self._get_frame_shape(frame)
            video_shape = self._get_video_shape(frame)
            self.encoder = _VideoEncoder(self.path, 
                                         frame_shape, 
                                         video_shape, 
                                         self.in_fps, 
                                         self.out_fps)

            self.metadata['encoder_version'] = self.encoder.version_info
            self.metadata['video_info'] = self.encoder.video_info
        
        try:
            self.encoder.capture_frame(frame)

        except RuntimeError as e:
            LOG.exception('Tried to pass invalid video frame, '
                        f'marking as broken: {self.path}')
            self.metadata['exc'] = str(e)
            self.broken = True
        else:
            self.empty = False

    def __del__(self):
        if self.enabled:
            print('WARNING: The _VideoRecorder is not closed, please '
                    'ensure you have closed the environment before the '
                    'program ended `env.close()`')
            self.close()


class VideoRecorder(MonitorToolChain):
    video_ext    = 'video.mp4'
    metadata_ext = 'metadata.json'
    video_suffix = 'ep{episodes:06d}.{start_steps}-{steps}'
    def __init__(self, root_dir:  str = None,
                       prefix:    str = None,
                       width:     int = None,
                       height:    int = None,
                       fps:       int = None,
                       env_fps:   int = None,
                       interval:  int = None,
                       metadata: dict = None,
                       force:    bool = True):
        '''This monitor tool records the environment screens.

        Args:
            root_dir (str, optional): Root path to store video recordings. 
                Defaults to None.
            prefix (str, optional): Video recording file prefix. Defaults 
                to None.
            width (int, optional): Video record width. Defaults to None.
            height (int, optional): Video record height. Defaults to None.
            fps (int, optional): Video record FPS. Defaults to None.
            env_fps (int, optional): Environment's FPS. Defaults to None.
            interval (int, optional): Record interval (episodes). It can be
                an int or a callable function which has a signature:
                    `def func(episode_number: int) -> bool`
                Return True if you want to record this number of episode,
                False otherwise. None for built in capped cubic video 
                schedule. Defaults to None.
            metadata (dict, optional): Metadata to write in additional json 
                file. Defaults to None.
            force (bool, optional): Force to save recordings. If False,
                it raises an exception if root_dir is not empty. Defaults 
                to True.

        Raises:
            RuntimeError: `interval` is not a callable object.
        '''        
        super().__init__()

        # Default callback
        if interval is None:
            schedule = VideoRecorder.capped_cubic_video_schedule
        elif isinstance(interval, int):
            schedule = lambda ep: (ep+1)%interval==0
        elif not callable(interval):
            raise RuntimeError('You must provide a function or int for '
                f'`interval` not {type(interval)}: {interval}')

        self._root_dir  = root_dir
        self._prefix    = prefix
        self._suffix    = self.video_suffix
        self._video_ext = self.video_ext
        self._width     = width
        self._height    = height
        self._fps       = fps
        self._env_fps   = env_fps
        self._meta_ext  = self.metadata_ext
        self._metadata  = metadata
        self._schedule  = schedule
        self._force     = force

        self._is_reset  = False # Whether the env is reset
        self._env       = None
        self._enabled   = True
        self._recorder  = None
        self._monitor   = None
        self._stats     = None

        # initialized in `set_monitor`

    def set_monitor(self, monitor: 'Monitor'):
        self._monitor = monitor
        self._stats   = monitor.stats
        self._env     = monitor.env
        # Default save root: {monitor.root_dir}/videos/
        self._root_dir = (self._root_dir 
            or os.path.join(self._monitor.root_dir, 'videos/'))
        # Default prefix: {monitor.prefix}
        self._prefix = (self._prefix or self._monitor.prefix)
        # Raise an exception if the path is not empty
        if not self._force:
            if not self._path_is_empty(self._root_dir):
                raise RuntimeError('Try to write to non-empty video '
                    f'directory {self._root_dir}')
    
    @staticmethod
    def capped_cubic_video_schedule(episode):
        if episode < 1000:
            return int(round(episode ** (1./3))) ** 3 == episode
        else:
            return episode % 1000 == 0
    
    @property
    def stats(self):
        return self._monitor.stats

    @property
    def need_record(self):
        '''Whether we should record this episode'''
        return self._schedule(self.stats.episodes)

    def close(self):
        '''Close VideoRecorder'''
        if not self._enabled:
            return
        self._close_and_save_recorder()
        # disable VideoRecorder
        self._enabled = False

    def _before_step(self, act):
        if not self._enabled:
            return act

        # If the env is reset, create a new recorder
        if self._is_reset:
            self._new_recorder()
            # capture the first frame
            self._recorder.capture_frame()
            self._is_reset = False
        return act

    def _after_step(self, *data):
        if not self._enabled:
            return data
        self._recorder.capture_frame()
        return data

    def _before_reset(self, **kwargs):
        if not self._enabled:
            return kwargs
        
        self._close_and_save_recorder()
        return kwargs

    def _after_reset(self, obs):
        if not self._enabled:
            return obs
        self._is_reset = True
        return obs

    def _create_path(self, path):
        os.makedirs(path, exist_ok=True)

    def _get_temp_path(self, dir, prefix, suffix):
        with tempfile.NamedTemporaryFile(dir=dir, prefix=prefix,
                suffix=suffix, delete=True) as f:
            path = str(f.name)
        return path

    def _make_path(self, root_dir, prefix, suffix, ext):
        paths = []
        if prefix:
            paths.append(str(prefix))
        if suffix:
            paths.append(str(suffix))
        paths.append(str(ext))

        filename = '.'.join(paths)
        path = os.path.join(root_dir, filename)
        abspath = os.path.abspath(path)
        return abspath.format(**self.stats)

    def _close_and_save_recorder(self):
        if not self._recorder:
            return
        if not self._recorder.functional:
            self._recorder.close()
            return

        video_path = self._make_path(
            root_dir = self._root_dir,
            prefix   = self._prefix,
            suffix   = self._suffix,
            ext      = self._video_ext
        )
        meta_path  = self._make_path(
            root_dir = self._root_dir,
            prefix   = self._prefix,
            suffix   = self._suffix,
            ext      = self._meta_ext
        )

        video_relpath = os.path.relpath(video_path)
        meta_relpath  = os.path.relpath(meta_path)

        # update metadata
        episode_metadata = {
            'episode':         self.stats.episodes,
            'start_steps':     self.stats.start_steps,
            'end_steps':       self.stats.steps,
            'episode_length':  self.stats.steps - self.stats.start_steps,
            'episode_rewards': self.stats.rewards
        }

        self._recorder.metadata['episode_info'] = episode_metadata
        self._recorder.write_metadata()
        # close recorder after writintg metadata
        self._recorder.close()
        # create save paths
        self._create_path(os.path.dirname(video_path))
        self._create_path(os.path.dirname(meta_path))

        # copy tempfiles to the specified locations
        if os.path.isfile(self._recorder.meta_path):
            os.rename(self._recorder.meta_path, meta_path)
            LOG.debug(f'Saving metadata to: {meta_path}')
        else:
            LOG.warning(f'Metadata not found: {meta_path}')

        if os.path.isfile(self._recorder.path):
            os.rename(self._recorder.path, video_path)
            LOG.debug(f'Saving video to: {video_path}')
        else:
            if self._recorder.broken:
                LOG.warning('Failed to save video, the VideoRecorder is broken,'
                            f'for more info: {meta_path}')
            else:
                LOG.error('Failed to save video, missing tempfile, '
                            f'for more info: {meta_path}')

    def _new_recorder(self):
        '''Create a new video recorder
        '''

        temppath = self._make_path(
            root_dir = self._root_dir,
            prefix   = self._prefix,
            suffix   = None,
            ext      = 'ep{episodes:06d}.'
        )

        root_dir   = os.path.dirname(temppath)
        prefix     = os.path.basename(temppath)
        video_path = None
        meta_path  = None

        if self.need_record:
            self._create_path(root_dir)
            video_path = self._get_temp_path(root_dir, prefix, '.mp4')
            meta_path  = self._get_temp_path(root_dir, prefix, '.json')

        self._recorder = _VideoRecorder(
            env       = self._env,
            path      = video_path,
            meta_path = meta_path,
            metadata  = self._metadata,
            width     = self._width,
            height    = self._height,
            in_fps    = self._env_fps,
            out_fps   = self._fps,
            enabled   = self.need_record
        )
    
    def _path_is_empty(self, path):
        if not path:
            return True
        
        return ((not os.path.isdir(path))
            or (len(os.listdir(path)) == 0))



# --- built in ---
import os
import sys
import time
import random
import datetime

# --- 3rd party ---
import gym
import cloudpickle

import numpy as np
import tensorflow as tf

# --- my module ---
from unstable_baselines.lib import utils as ub_utils
from unstable_baselines.lib import prob as ub_prob
from unstable_baselines.lib import patch as ub_patch

# === Common nets ===

class Identity(tf.keras.Model):
    '''Do nothing'''
    def __init__(self, **kwargs):
        super().__init__(**kwargs)
    def call(self, inputs, training=True):
        return inputs

class Constant(tf.keras.Model):
    '''Return a constant'''
    def __init__(self, constant, **kwargs):
        super().__init__(**kwargs)
        self.constant = constant
    def call(self, inputs, training=True):
        # broadcast shape to batch shape (b, *constant.shape)
        batch_shape = tf.shape(inputs)[0]
        value = tf.expand_dims(self.constant, axis=0)
        return tf.repeat(value, batch_shape, axis=0)

class MlpNet(tf.keras.Model):
    '''MLP feature extractor'''
    def __init__(self, mlp_units=[64, 64], **kwargs):
        super().__init__(**kwargs)

        layers = [tf.keras.layers.Flatten()]
        for unit in mlp_units:
            layers.extend([
                tf.keras.layers.Dense(unit),
                ub_patch.ReLU()
            ])
        
        self._model = tf.keras.Sequential(layers)

    def call(self, inputs, training=True):
        return self._model(inputs, training=training)

class NatureCnn(tf.keras.Model):
    def __init__(self, **kwargs):
        '''Nature CNN originated from 
        "Playing Atari with Deep Reinforcement Learning"
        '''        
        super().__init__(**kwargs)

        self._model = tf.keras.Sequential([
            tf.keras.layers.Conv2D(32, 8, 4),
            ub_patch.ReLU(),
            tf.keras.layers.Conv2D(64, 4, 2),
            ub_patch.ReLU(),
            tf.keras.layers.Conv2D(64, 3, 1),
            ub_patch.ReLU(),
            tf.keras.layers.Flatten(),
            tf.keras.layers.Dense(512),
            ub_patch.ReLU()
        ])

    def call(self, inputs, training=True):
        return self._model(inputs, training=training)

# === Complex nets ===

class AwesomeNet(tf.keras.Model):
    def __init__(self, input_space: gym.Space, 
                       force_mlp:   bool = False, 
                       mlp_units:   list = [64, 64], 
                       **kwargs):
        '''AwesomeNet automatically handles nested spaces and input 
        tensors. AwesomeNet creates either NatureCnn or MlpNet for 
        each observation space according to its type if `force_mlp`
        is False. Otherwise, all observations forwards into a MlpNet.
        '''
        super().__init__(**kwargs)
        self.force_mlp = force_mlp
        self.mlp_units = mlp_units
        # flatten nested input space to a tuple of spaces
        self._flat_spaces = ub_utils.flatten_space(input_space)
        assert len(self._flat_spaces) > 0
        # create models for each space
        self._models = [
            self.create_model(space)
            for space in self._flat_spaces
        ]
        
    def call(self, inputs, training=True):
        '''Forward networks'''
        flat_inputs = tf.nest.flatten(inputs)
        # assert len(flat_inputs) == len(self._models)
        outputs = []
        for input, model in zip(flat_inputs, self._models):
            x = model(input, training=training)
            outputs.append(x)
        return tf.concat(outputs, axis=1)

    def create_model(self, space: gym.Space):
        '''Create model by space'''
        use_cnn = (ub_utils.is_image_space(space)
                    and not self.force_mlp)
        if use_cnn:
            return self.create_cnn()
        else:
            return self.create_mlp()

    def create_cnn(self):
        '''Create nature cnn'''
        return NatureCnn()

    def create_mlp(self):
        '''Create mlp net'''
        return MlpNet(self.mlp_units)

# === Policies ===

class CategoricalPolicyNet(tf.keras.Model):
    '''Categorical policy for discrete action space'''
    support_spaces = [gym.spaces.Discrete]
    def __init__(self, action_space, **kwargs):
        '''Create categorical policy net

        Args:
            action_space (gym.spaces.Spaces): Action space, must be
                one of the types listed in `support_spaces`
        '''
        super().__init__(**kwargs)

        if not isinstance(action_space, tuple(self.support_spaces)):
            raise ValueError(f'{type(self)} does not support '
                f'action spaces of type `{type(action_space)}`')

        self.action_space = action_space
        self._model = None

    def build(self, input_shape):
        self._model = self.create_model()
    
    def call(self, inputs, training=True):
        '''Forward network

        Args:
            inputs (tf.Tensor): Expecting a latent vector in shape
                (b, latent), tf.float32
            training (bool, optional): Training mode. Defaults to True.

        Returns:
            Categorical: A categorical distribution
        '''        
        return ub_prob.Categorical(self._model(inputs, training=training))

    def create_model(self):
        '''Override this method to customize model arch'''
        return tf.keras.Sequential([
            tf.keras.layers.Dense(self.action_space.n)
        ])

class DiagGaussianPolicyNet(tf.keras.Model):
    '''Tanh squashed diagonal Gaussian mixture policy net 
    with state-dependent covariance
    '''
    support_spaces = [gym.spaces.Box]
    def __init__(self, action_space, squash=False, **kwargs):
        '''Create diagonal gaussian policy net

        Args:
            action_space (gym.spaces): Action space, must be
                one of the types listed in `support_spaces`
            squash (bool): Tanh squashing. Default to False.
        '''
        super().__init__(**kwargs)

        if not isinstance(action_space, tuple(self.support_spaces)):
            raise ValueError(f'{type(self).__name__} does not '
                f'suprt action space of type `{type(action_space)}`')

        self.action_space = action_space
        self.action_dims  = int(np.prod(action_space.shape))
        self.action_shape = action_space.shape
        self.squash       = squash

        self._mean_model   = None
        self._logstd_model = None

    def build(self, input_shape):
        '''Build model
        '''
        self._mean_model   = self.create_mean_model()
        self._logstd_model = self.create_logstd_model()

    def call(self, inputs, training=True):
        '''Forward network

        Args:
            inputs (tf.Tensor): Expecting a latent vector in shape
                (b, latent), tf.float32
            training (bool, optional): Training mode. Defaults to True.

        Returns:
            MultiNormal: A multi variate gaussian distribution
        '''
        # forward model
        mean   = self._mean_model(inputs, training=training)
        logstd = self._logstd_model(inputs, training=training)
        std    = tf.math.softplus(logstd) + 1e-5
        # reshape as action space shape (-1 = batch dim)
        output_shape = [-1] + list(self.action_shape)
        mean = tf.reshape(mean, output_shape)
        std  = tf.reshape(std, output_shape)
        # create multi variate gauss dist with tah squashed
        distrib = ub_prob.MultiNormal(mean, std)
        if self.squash:
            distrib = ub_prob.Tanh(distrib)
        return distrib

    def create_mean_model(self):
        '''Mean/Loc model
        Override this method to customize model arch

        Returns:
            tf.keras.Model: mean/log model
        '''        
        return tf.keras.Sequential([
            tf.keras.layers.Dense(self.action_dims)
        ])

    def create_logstd_model(self):
        '''Log-std model
        Override this method to customize model arch

        Returns:
            tf.keras.Model: log-std model
        '''        
        return tf.keras.Sequential([
            tf.keras.layers.Dense(self.action_dims)
        ])

class PolicyNet(tf.keras.Model):
    support_spaces = [gym.spaces.Box, gym.spaces.Discrete]
    #TODO: support other spaces: MultiDiscrete, Dict, Tuple
    def __init__(self, action_space, squash=False, net=None, **kwargs):
        '''Base Policy net
        Override `create_model` or `create_*_policy` method to customize
        model architecture

        Args:
            action_space (gym.Space): The action space.
            squash (bool, optional): Apply Tanh squash for continuous (Box) 
                action space. Defaults to False.
            net (tf.keras.Model, optional): Base network, feature extractor.
                Defaults to None.

        Raises:
            ValueError: if the given action space type does not contain in the 
                `support_spaces` list
        '''
        super().__init__(**kwargs)
        # check if space supported
        if not isinstance(action_space, tuple(self.support_spaces)):
            raise ValueError(f'{type(self)} does not '
                f'support the action space of type {type(action_space)}')
        
        self.action_space = action_space
        self.squash  = squash
        self._model  = net
        self._policy = None
    
    def build(self, input_shape):
        # create feature extraction model
        if self._model is None:
            self._model = Identity()
        # create policy
        if isinstance(self.action_space, gym.spaces.Discrete):
            self._policy = self.create_categorical_policy()
        elif isinstance(self.action_space, gym.spaces.Box):
            self._policy = self.create_gaussian_policy()
        else:
            raise ValueError(f'{type(self)} dose not '
                f'support the action space of type {type(action_space)}')

    def call(self, inputs, training=True):
        '''Forward network

        Args:
            inputs (tf.Tensor): Input tensors, shape (b, *)
            training (bool, optional): Training mode. Defaults to True.

        Returns:
            Distribution: Action distributions.
        '''
        x = self._model(inputs, training=training)
        return self._policy(x, training=training)
    
    def create_categorical_policy(self):
        '''Policy for Discrete action spaces'''
        return CategoricalPolicyNet(self.action_space)

    def create_gaussian_policy(self):
        '''Policy for Box action space'''
        return DiagGaussianPolicyNet(self.action_space, self.squash)


# === Value networks ===

class ValueNet(tf.keras.Model):
    def __init__(self, net=None, **kwargs):
        '''Base value net

        Args:
            net (tf.keras.Model, optional): Base network, feature extractor. 
                Defaults to None.
        '''        
        super().__init__(**kwargs)
        self._model = net
        self._value = None

    def build(self, input_shape):
        # create empty net
        if self._model is None:
            self._model = Identity()
        self._value = self.create_value_model()

    def call(self, inputs, training=True):
        '''Forward value net

        Args:
            inputs (tf.Tensor): Input tensors, shape (b, *).
            training (bool, optional): Training mode. Defaults to True.

        Returns:
            tf.Tensor: value predictions, (Default) shape (b, 1)
        '''
        x = self._model(inputs, training=training)
        return self._value(x, training=training)
    
    def create_value_model(self):
        '''Create value model
        Override this method to customize model arch

        Returns:
            tf.keras.Model: model
        '''
        return tf.keras.layers.Dense(1)

class MultiHeadValueNets(tf.keras.Model):
    def __init__(self, n_heads: int=2, nets=None, **kwargs):
        '''Base multi head value net

        Args:
            n_heads (int, optional): Number of heads. Defaults to 2.
            nets (tf.keras.Model, list, optional): Base networks, 
                feature extractors. Can be a tf.keras.Model or a list
                of tf.keras.Model. Note that if only a tf.keras.Model 
                was provided, it is shared across all heads. The count 
                of tf.keras.Model of the list must match to `n_heads`. 
                Defaults to None.
        '''
        super().__init__(**kwargs)
        self.n_heads = n_heads
        self._models = nets
        self._values = None

    def build(self, input_shape):
        if self._models is None:
            self._models = [
                Identity()
                for n in range(self.n_heads)
            ]
        elif isinstance(self._models, (list, tuple)):
            pass
        else:
            # share net, copy the same net multiple times
            self._models = [
                self._models
                for n in range(self.n_heads)
            ]
        assert len(self._models) == self.n_heads
        # create value nets
        self._values = [
            self.create_value_model()
            for i in range(self.n_heads)
        ]

    def call(self, inputs, index=None, axis=0, training=True):
        '''Forward all critics

        Args:
            inputs (tf.Tensor): Expecting a batch latents with shape
                (b, latent), tf.float32
            index (int): index of value net to forward. If None, forwards
                all nets. Defaults to None.
            axis (int): axis to stack values.
            training (bool, optional): Training mode. Defaults to True.

        Returns:
            tf.Tensor: Value predictions. (Default) shape (n_heads, b, 1)
        '''
        if index is None:
            # forward all value nets and stack results
            return tf.stack([
                self._values[n](
                    self._models[n](inputs, training=training), 
                    training=training
                )
                for n in range(self.n_heads)
            ], axis=axis)
        else:
            try:
                x = self._models[index](inputs, training=training)
                return self._values[index](x, training=training)
            except IndexError:
                raise IndexError(f'Index out of range, n_head={self.n_heads}'
                    f' got index={index}')

    def create_value_model(self):
        '''Create model
        Override this method to customize model arch

        Returns:
            tf.keras.Model: model
        '''        
        return tf.keras.layers.Dense(1)


# --- built in ---
import os
import sys
import time
import logging

from logging.config import dictConfig
from collections import OrderedDict
from string import Formatter

from logging import CRITICAL, DEBUG, ERROR, FATAL, INFO, WARN, WARNING, getLevelName, NOTSET, addLevelName

LOG_FORMAT = '[%(asctime)s|%(threadName)s|%(levelname)s|%(name)s:%(filename)s:%(lineno)d]: %(message)s'
LOG_DATE_FORMAT = '%Y-%m-%d %H:%M:%S'

ROOT_NAME = __name__.split('.', 1)[0]
MODULE_PATH = __name__.rsplit('.', 1)[0]
formatter_module = __import__(MODULE_PATH+'.formatter', fromlist=['.'])

MultilineStringFormatter = getattr(formatter_module, 'MultilineStringFormatter')


COLORED_LOG_FORMAT = '$TIME%(asctime)s$RESET $THREAD%(threadName)s$RESET $LEVEL[%(levelname)s]$RESET $MODULE%(name)s:%(filename)s:%(lineno)d$RESET: $LEVEL%(message)s$RESET'


__all__ = [
    'CRITICAL', 'DEBUG', 'ERROR', 'FATAL', 'INFO', 'WARN', 'WARNING', 'getLevelName', 'NOTSET', 'addLevelName',
    'critical', 'debug', 'error', 'fatal', 'info', 'warn', 'warning', 'log',
    'getLogger', 'CardLogger'
]

def _exists(name):
    '''
    Check if a logger exists
    '''
    return name in logging.Logger.manager.loggerDict


def exists(name=None, prefix=ROOT_NAME):
    if name is not None:
        name = '.'.join([prefix, name])
    else:
        name = prefix

    return _exists(name)


def getLogger(name=None, level=None, prefix=ROOT_NAME):
    if name is not None:
        name = '.'.join([prefix, name])
    else:
        name = prefix


    # if logger not exist, and no level specified, use default level (DEBUG)
    if (level is None) and (not _exists(name)):
        level = 'DEBUG' 
        

    logger = logging.getLogger(name)

    if level is not None:
        logger.setLevel(logging.getLevelName(level))

    return logger


def log(level, msg, *args, **kwargs):
    getLogger()._log(level, msg, args, **kwargs)

def critical(msg, *args, **kwargs):
    getLogger()._log(CRITICAL, msg, args, **kwargs)

def fatal(msg, *args, **kwargs):
    getLogger()._log(CRITICAL, msg, args, **kwargs)

def error(msg, *args, **kwargs):
    getLogger()._log(ERROR, msg, args, **kwargs)

def exception(msg, *args, exc_info=True, **kwargs):
    getLogger()._log(ERROR, msg, args, exc_info=exc_info, **kwargs)

def warning(msg, *args, **kwargs):
    getLogger()._log(WARNING, msg, args, **kwargs)

def warn(msg, *args, **kwargs):
    getLogger()._log(WARN, msg, args, **kwargs)

def debug(msg, *args, **kwargs):
    getLogger()._log(DEBUG, msg, args, **kwargs)

def info(msg, *args, **kwargs):
    getLogger()._log(INFO, msg, args, **kwargs)


class CardLogger(logging.Logger):
    
    #'======= Epoch 5/10 ======='
    #'|------- Training -------|'
    #'| loss: 100.0            |'
    #'| entropy: 1132121.456   |'
    #'|--------- Eval ---------|'
    #'| loss: 2121.17455       |'
    #'=========================='

    _alignment_abbreviation = {'l': 'left', 'c': 'center', 'r': 'right'}

    def __init__(self, name, level=logging.NOTSET):
        super().__init__(name, level)

        self.header = None
        self.groups = OrderedDict()
        self.groups['None'] = []
        self.current_group = 'None'

    def set_header(self, name=None):
        if not isinstance(name, str):
            raise ValueError('Header name must be a str type')
        self.header = name

    def subgroup(self, name=None):

        if name is None:
            name = 'None'

        if not isinstance(name, str):
            raise ValueError('Group name must be str type')

        self.current_group = name

    def add_row(self, *args, fmt=None, align='left', **kwargs):
        if not self.current_group in self.groups:
            self.groups[self.current_group] = []

        if align in self._alignment_abbreviation:
            align = self._alignment_abbreviation[align]

        assert align in ['left', 'center', 'right']

        self.groups[self.current_group].append( (args, kwargs, fmt, False, align, False) )

    def add_rows(self, *args, fmt=None, align='left', **kwargs):
        if not self.current_group in self.groups:
            self.groups[self.current_group] = []

        if align in self._alignment_abbreviation:
            align = self._alignment_abbreviation[align]
        
        assert align in ['left', 'center', 'right']

        self.groups[self.current_group].append( (args, kwargs, fmt, True, align, False) )

    def add_line(self):
        
        # args, kwargs, fmt, Multiline, align, horizontal line
        self.groups[self.current_group].append( (None, None, None, False, None, True) )

    def clear(self):
        self.header = None
        self.groups = OrderedDict()
        self.groups['None'] = []
        self.current_group = 'None'

    def _create_header(self, width):
        if self.header is None:
            header = '=' * width

        else:
            half = (width - len(self.header)) // 2 - 1
            header = ' '.join(['=' * half, self.header, '=' * half])

        return header

    def _create_tail(self, width):
        return '=' * width

    def _create_subheader(self, name, width):

        remain = width - len(name)
        l_half = remain // 2
        r_half = remain - l_half

        l_half -= 2
        r_half -= 2

        subheader = ''.join(['|', '-'*l_half, ' ', name, ' ', '-'*r_half, '|'])

        return subheader

    def _create_row(self, string, align, width):
        remain = width - len(string) - 4

        if align == 'center':
            l = remain//2
            r = remain - l
        elif align == 'left':
            l = 0
            r = remain
        else: #right
            r = 0
            l = remain

        return ''.join(['| ', ' '*l, string, ' '*r, ' |'])

    def _create_line(self, width):
        remain = width - 2 # | + ----- + |

        return ''.join([ '|', '-'*remain, '|' ])

    def _format_contents(self):
        '''
        Formatting contents
        '''
        
        group_str = OrderedDict()

        max_length = 0

        # compute max row length
        for group, rows in self.groups.items():

            if group not in group_str:
                group_str[group] = []

            for row in rows:
                args = row[0]
                kwargs = row[1]
                fmt = row[2]
                multi_line = row[3]
                align = row[4]
                line = row[5]


                if line:
                    # string, align, horizontal line
                    group_str[group].append((None, None, True))
                    continue

                if not multi_line:

                    if fmt is None:
                        if len(args) > 0:
                            fmt = '{}'
                            if len(args) > 1:
                                fmt += ': ' + ', '.join(['{}' for _ in range(len(args[1:]))])
                        else:
                            fmt = ''

                    string = fmt.format(*args, **kwargs)

                    group_str[group].append( (string, align, False) )
                
                    max_length = max(max_length, len(string))

                else:
                    # print multi line

                    if fmt is None:
                        if len(args) > 0:
                            fmt = '{}'
                            if len(args) > 1:
                                fmt += ': ' + ', '.join(['{}' for _ in range(len(args[1:]))])
                        else:
                            fmt = ''

                    strings = MultilineStringFormatter().format(fmt, *args, **kwargs)

                    for line in strings:
                        group_str[group].append( (line, align, False) )
                        max_length = max(max_length, len(line))

        # calculate max length
        max_width = max_length + len('|  |')

        # calculate max sub header length
        for group in self.groups.keys():
            group_width = len(group) + len('|-  -|')

            max_width = max(max_width, group_width)

        if self.header is not None:
            header_width = len(self.header) + len('=====  =====')

            max_width = max(max_width, header_width)

            if (max_width - len(self.header)) % 2 > 0:
                max_width += 1


        # formatting message
        messages = []
        messages.append(self._create_header(max_width))

        for group, contents in group_str.items():

            if len(contents) == 0:
                continue

            if group is not 'None':
                messages.append(self._create_subheader(group, max_width))

            for content in contents:
                string = content[0]
                align = content[1]
                line = content[2]
                # horizontal line
                if line:
                    messages.append(self._create_line(max_width))
                else:
                    messages.append(self._create_row(string, align, max_width))

        messages.append(self._create_tail(max_width))
            
        return messages
    

    def flush(self, level=logging.INFO, 
                    tostring=None, 
                    tolist=None, 
                    new_line=True,
                    contents=None):
        '''
        Flush content
        
        args:
            tostring: convert to a single string
            tolist: convert to a list of strings
            new_line: add a new line after the contents
            contents: a str or a list of str
        '''


        if isinstance(level, str):
            level = logging.getLevelName(level)

        if not isinstance(level, int):
            if logging.raiseExceptions:
                raise TypeError('level must be an integer or string')
            else:
                return
        
        if self.isEnabledFor(level):
        
            if contents is not None:
            
                assert isinstance(contents, (str, list)), 'The contents must be a str or a list of str'
                
                if isinstance(contents, str):
                    contents = contents.split('\n')
            else:
                contents = self._format_contents()
                
            
            assert isinstance(contents, list)
            
            # add a new line
            if new_line:
                contents.append('')
                
                
            assert not (tostring and tolist), 'tostring and tolist, only one can be True'
            
            # convert to list
            if tolist:
                self.clear()
                return contents
                
            # convert to string
            if tostring:
                contents = '\n'.join(contents)
                self.clear()
                return contents
                
            # output to log
            for c in contents:
                self._log(level, c, None)


            '''
            self._log(level, self._create_header(max_width), None)

            for group, contents in group_str.items():

                if len(contents) == 0:
                    continue

                if group is not 'None':

                    self._log(level, self._create_subheader(group, max_width), None)

                for content in contents:
                    string = content[0]
                    align = content[1]
                    line = content[2]
                    # horizontal line
                    if line:
                        self._log(level, self._create_line(max_width), None)
                    else:
                        self._log(level, self._create_row(string, align, max_width), None)

            self._log(level, self._create_tail(max_width), None)
            self._log(level, '', None)
            '''

        self.clear()
        
        if tostring:
            return ''
        if tolist:
            return []


class LoggingConfig:
    def __init__(self, filename=None, level='INFO', colored=True, **kwargs):

        '''
        args:
            filename: logging file name. It will not save to file if it is 'None'
            level: logging level on console. The logging level saving to file is always set to 'DEBUG'
            colored: colored console log
        kwargs:
            reset: reset loggers. Setting this to True will disable all existed loggers
        '''
        
        console_formatter = ROOT_NAME+'-colored' if colored else ROOT_NAME

        output_to_file = not (filename is None)
        
        if output_to_file:

            path = os.path.dirname(filename)
            if path:
                os.makedirs(path, exist_ok=True)

        config = {
            'version': 1,
            'formatters': {
                ROOT_NAME+'-colored': {
                    '()': MODULE_PATH+'.ColoredFormatter',
                    'format': COLORED_LOG_FORMAT,
                    'datefmt': LOG_DATE_FORMAT,
                },
                ROOT_NAME: {
                    'format': LOG_FORMAT,
                    'datefmt': LOG_DATE_FORMAT,
                }
            },
            'handlers': {
                'console': {
                    'level': level,
                    'class': 'logging.StreamHandler',
                    'stream': 'ext://sys.stdout',
                    'formatter': console_formatter,
                },
            },
            'loggers': {
                ROOT_NAME: {
                    'handlers': ['console'] ,
                    'level': 'DEBUG',
                    'propagate': False,
                }
            }
        }

        if output_to_file:
            config['handlers'][filename] = {
                'level': 'DEBUG',
                'filename': filename,
                'class': 'logging.FileHandler',
                'formatter': ROOT_NAME
            }

            config['loggers'][ROOT_NAME]['handlers'] += [filename]


        self.config = config



    def apply(self, reset=False):
        '''
        Apply configurations

        reset: reset logger. Setting this to True will disable all existing loggers.
        '''


        self.config['disable_existing_loggers'] = reset
        dictConfig(self.config)

    @classmethod
    def use(cls, **kwargs):
        reset = kwargs.pop('reset', False)

        conf = cls(**kwargs)
        conf.apply(reset=reset)

        return conf


logging.setLoggerClass(CardLogger)
